#!/usr/bin/perl -w

package Alignments;
use Exit(exit_nicely);

$common_substitutions = 'X:-,B:-,O:-';

# These scales are set in use_*() and calc_entropy() subroutines to
#   1) bias the properties E1 to E5 when necessary
#   2) generate entropies from the interval 0-1 regardless of the method used.
#
@scaleS = ();

$aa_types = 
{
    'A' => 'Hydrophobic',
    'R' => 'Basic',
    'N' => 'Polar',
    'D' => 'Acidic',
    'C' => 'Polar',
    'Q' => 'Polar',
    'E' => 'Acidic',
    'G' => 'Polar',
    'H' => 'Basic',
    'I' => 'Hydrophobic',
    'L' => 'Hydrophobic',
    'K' => 'Basic',
    'M' => 'Hydrophobic',
    'F' => 'Hydrophobic',
    'P' => 'Hydrophobic',
    'S' => 'Polar',
    'T' => 'Polar',
    'W' => 'Hydrophobic',
    'Y' => 'Polar',
    'V' => 'Hydrophobic'
};

$aa_colors =
{
    'Hydrophobic' => { fg => '000000', bg => 'ffffff' },
    'Polar'       => { fg => '008300', bg => 'ffffff' },
    'Acidic'      => { fg => 'ea0000', bg => 'ffffff' },
    'Basic'       => { fg => '008bea', bg => 'ffffff' },
    'default'     => { fg => '000000', bg => 'ffffff' },
};

#---------------------------------------------------------------


sub select_total_entropy_method
{
    my ($method) = @_;

    if ( !defined($Vectorlib::nprops) ) { exit_nicely("Have you forgot to include vectorlib.pm in your script?\n") }

    if ( $method eq 'unscaled_sum' ) { use_unscaled_sum() }
    elsif ( $method eq 'scaled_sum' ) { use_scaled_sum() }
    elsif ( $method eq 'unscaled_max' ) { use_unscaled_max() }
    elsif ( $method eq 'ori' ) { use_original_approach() }
    else { Exit::exit_nicely("The total entropy approach '$method' not recognised.\n") }
}

sub use_scaled_sum
{
    if ( !defined($Vectorlib::nprops) ) { exit_nicely("Have you forgot to include vectorlib.pm in your script?\n") }

    if ( $total_entropy_approach && $total_entropy_approach eq 'scaled_sum' ) { return }
    $total_entropy_approach = 'scaled_sum';

    # These scales are used for biasing the E1-5 property space in
    #   such a way, that the relative entropies have equal ranges 0-1.
    #
    @scaleS = ();
    for my $i (0..$Vectorlib::nprops-1) { $scaleS[$i] = 1; }
    $entropy_limits = { least=>1, largest=>1 };

    # Determine for all properties what is their maximum and minimum possible
    #   value. (Runs over the 20 amino acids.)
    my @max = ();
    my @min = ();
    for my $aa (keys %$aa_types)
    {
        my $hist = { $aa=>1 };
        my ($entrps,$avgs,$devs) = calc_entropy($hist);
        
        if ( ! scalar @max ) { (@max) = @$entrps }
        if ( ! scalar @min ) { (@min) = @$entrps }

        for (my $i=0; $i<scalar @$entrps; $i++)
        {
            if ( $max[$i] < $$entrps[$i] ) { $max[$i] = $$entrps[$i] }
            if ( $min[$i] > $$entrps[$i] ) { $min[$i] = $$entrps[$i] }
        }
    }
    if ( scalar @min != $Vectorlib::nprops ) { exit_nicely(sprintf "FIXME: %d vs. %d\n",scalar @min,$Vectorlib::nprops) }

    for my $i (0..$Vectorlib::nprops-1) { $scaleS[$i] = 1.0/$max[$i]; }
    my $min = -1;
    my $max = -1;
    my $min_aa = '';
    my $max_aa = '';
    for my $aa (keys %$aa_types)
    {
        my $hist = { $aa=>1 };
        my ($entrps,$avgs,$devs,$total) = calc_entropy($hist);

        #print STDERR "$aa: "; for my $i (0..4) { printf STDERR " %.2f",$$entrps[$i]; } printf STDERR "  %.2f\n",$total;

        if ( $min==-1 || $min>$total ) { $min = $total; $min_aa = $aa; }
        if ( $max==-1 || $max<$total ) { $max = $total; $max_aa = $aa; }
    }
    #print STDERR "Using '$total_entropy_approach' approach. Min entropy is $min ($min_aa), max is $max ($max_aa)\n";

    $entropy_limits = { least=>$min, largest=>$max };
}


sub use_unscaled_sum
{
    if ( !defined($Vectorlib::nprops) ) { exit_nicely("Have you forgot to include vectorlib.pm in your script?\n") }

    if ( $total_entropy_approach && $total_entropy_approach eq 'unscaled_sum' ) { return }

    $total_entropy_approach = 'unscaled_max';   # First get the unscaled maximum
    @scaleS = ();
    for my $i (0..$Vectorlib::nprops-1) { $scaleS[$i] = 1; }
    $entropy_limits = { least=>1, largest=>1 };
    my $max = -1;
    for my $aa (keys %$aa_types)
    {
        my $hist = { $aa=>1 };
        my ($entrps,$avgs,$devs,$total) = calc_entropy($hist);

        if ( $max==-1 || $max<$total ) { $max = $total; }
    }

    # Although it is called "unscaled", there always is some scaling involved to
    #   get the entropies from the interval <0,1>
    #
    for my $i (0..$Vectorlib::nprops-1) { $scaleS[$i] = 1/$max; }
    $total_entropy_approach = 'unscaled_sum';

    $max = -1;
    my $min = -1;
    my $min_aa = '';
    my $max_aa = '';
    for my $aa (keys %$aa_types)
    {
        my $hist = { $aa=>1 };
        my ($entrps,$avgs,$devs,$total) = calc_entropy($hist);

        if ( $min==-1 || $min>$total ) { $min = $total; $min_aa = $aa; }
        if ( $max==-1 || $max<$total ) { $max = $total; $max_aa = $aa; }
    }

    $entropy_limits = { least=>$min, largest=>$max };
}

sub use_unscaled_max
{
    if ( !defined($Vectorlib::nprops) ) { exit_nicely("Have you forgot to include vectorlib.pm in your script?\n") }

    if ( $total_entropy_approach && $total_entropy_approach eq 'unscaled_max' ) { return }
    $total_entropy_approach = 'unscaled_max';

    @scaleS = ();
    for my $i (0..$Vectorlib::nprops-1) { $scaleS[$i] = 1; }
    $entropy_limits = { least=>1, largest=>1 };

    my $min = -1;
    my $max = -1;
    my $min_aa = '';
    my $max_aa = '';
    for my $aa (keys %$aa_types)
    {
        my $hist = { $aa=>1 };
        my ($entrps,$avgs,$devs,$total) = calc_entropy($hist);

        if ( $min==-1 || $min>$total ) { $min = $total; $min_aa = $aa; }
        if ( $max==-1 || $max<$total ) { $max = $total; $max_aa = $aa; }
    }

    # Although it is called "unscaled", there always is some scaling involved to
    #   get the entropies from the interval <0,1>
    #
    for my $i (0..$Vectorlib::nprops-1) { $scaleS[$i] = 1/$max; }
    $entropy_limits = { least=>$min/$max, largest=>$max/$max };
}


sub use_original_approach
{
    if ( !defined($Vectorlib::nprops) ) { exit_nicely("Have you forgot to include vectorlib.pm in your script?\n") }

    $total_entropy_approach = 'ori';

    @scaleS = ();
    for my $i (0..$Vectorlib::nprops-1) { $scaleS[$i] = 1; }
    $entropy_limits = { least=>1, largest=>1 };
}


sub init_colors
{
    my ($aa_cols) = @_;

    if ( !$aa_cols || !scalar keys %$aa_cols ) { $aa_cols = $aa_colors; }

    for my $aa (keys %$aa_types,keys %$aa_cols)
    {
        my $type = exists($aa_types->{$aa}) ? $aa_types->{$aa} : $aa;
        for my $ground ('bg','fg')
        {
            $aa_cols->{$aa}->{$ground} = exists($aa_cols->{$type}->{$ground}) ? $aa_cols->{$type}->{$ground} : $aa_cols->{'default'}->{$ground};
            if ( !exists($aa_cols->{$aa}) || !exists($aa_cols->{$aa}->{$ground}) )
            {
                exit_nicely("FIXME: The amino acid type '$type' is not defined?\n");
            }

            if ( ref($aa_cols->{$aa}->{$ground}) ne 'ARRAY' )
            {
                my $r = hex(substr($aa_cols->{$aa}->{$ground},0,2)) / 255.;
                my $g = hex(substr($aa_cols->{$aa}->{$ground},2,2)) / 255.;
                my $b = hex(substr($aa_cols->{$aa}->{$ground},4,2)) / 255.;
                $aa_cols->{$aa}->{$ground} = [$r,$g,$b];
            }
        }
    }
    return $aa_cols;
}


# Returns pointer to array of sequences. Each entry of the array
#   is a hash with three keys: id, pad, seq.
#
sub parse_alignment
{
    my ($input_file,$options) = @_;

    if ( $options && ref($options) ne 'HASH' ) { Exit::exit_nicely("The parse_alignment interface has changed!!\n") }

    my $substitute   = '';
    my $check_sanity = 'strict';
    if ( $options )
    {
        if ( exists($$options{'substitute'}) ) { $substitute = $$options{'substitute'} }
        if ( exists($$options{'sanity_check'}) ) { $check_sanity = $$options{'sanity_check'} }  # Can be: strict|warn|none
    }

    my ($fh,$must_close);
    if ( ref($input_file) eq 'GLOB' ) 
    { 
        $fh = $input_file; 
    }
    elsif ( $input_file =~ /\.gz$/i )
    {
        open($fh,"zcat $input_file |") || exit_nicely("zcat $input_file |: $!");
    }
    else
    {
        open($fh,"<$input_file") || exit_nicely("$input_file: $!");
    }
    my @lines = ();
    while (my $line=<$fh>)
    {
        $line =~ s/\r?\n//g;
        push @lines,$line;
    }
    close($fh) unless !$must_close;

    my $seqs;
    if ( $lines[0]=~ /clustal/i ) 
    { 
        $seqs = read_clustal_aln(\@lines);
    }
    elsif ( $lines[0]=~ /^>/ )
    {
        $seqs = read_pir_aln(\@lines);
    }
    elsif ( $lines[0]=~ /^(\d+)\s+(\d+)$/ )
    {
        $seqs = read_table_aln(\@lines);
    }
    else
    {
        $seqs = read_oneline_aln(\@lines);
    }

    if ( $check_sanity ne 'none' )
    {
        check_alignment_sanity($seqs,{'mode'=>$check_sanity});
    }

    # Some non-standard letters will be replaced according to the given
    #   substitution patterns. See e.g. create-consensus-sequence -?.
    if ( $substitute )
    {
        my $replace = {};
        my @subs = split(/,/,$substitute);
        my $nsubs = scalar @subs;
        for (my $i=0; $i<$nsubs; $i++)
        {
            my ($ori,$new) = split(/:/,$subs[$i]);
            if ( !defined($ori) || !defined($new) ) 
            { 
                Exit::exit_nicely(
                        "Could not parse the substitution string '$substitute'\n",
                        "Expected a comma-separated list of values ORI:NEW, such as for example \"X:-,B:-,O:-\".\n",
                        ) 
            }
            $replace->{$ori} = $new;
        }

        for my $seq (@$seqs)
        {
            while ( my($ori,$repl) = each %$replace )
            {
                $seq->{seq} =~ s/$ori/$repl/g;
            }
        }
    }

    for my $seq (@$seqs) { $$seq{'from'} = 0; } # for use with aln-to-img 

    return $seqs;
}


# Sanity check: Are the sequences in multiple alignment of the same length?
sub check_alignment_sanity
{
    my ($aln,$options) = @_;
    my ($len, $prev_len);

    my $warn_only=0;
    if ( $options )
    {
        if ( exists($$options{'mode'}) && $$options{'mode'} eq 'warn' ) { $warn_only=1; } 
    }

    foreach my $seq (@$aln)
    {
        $len = length($seq->{seq});
        if ( $prev_len && $len!=$prev_len )
        {
            print STDERR 
                    "Sanity check failed: The sequences of the multiple alignment have different lengths:\n",
                    "   $prev_len vs $len for $seq->{id}.\n",
                    "Hint: Is this a multiple sequence alignment at all? Maybe you uploaded clustal *input* file instead?\n\n";

            Exit::exit_nicely("Sorry, giving up.\n") unless $warn_only;
        }
        $prev_len = $len;
    }
}

sub write_oneline_aln
{
    my ($seqs,$fname) = @_;

    my ($fh,$must_close);

    if ( ref($fname) eq 'GLOB' ) { $fh = $fname; }
    else { open($fh,">$fname") || exit_nicely("$fname: $!"); $must_close=1; }


    for (my $i=0; $i<scalar @$seqs; $i++)
    {
        print $fh $seqs->[$i]->{id};
        print $fh $seqs->[$i]->{pad};
        print $fh $seqs->[$i]->{seq};
        print $fh "\n";
    }

    if ( $must_close ) { close($fh); }
}


sub write_table_aln
{
    my ($seqs,$fname) = @_;

    my ($fh,$must_close);

    if ( ref($fname) eq 'GLOB' ) { $fh = $fname; }
    else { open($fh,">$fname") || exit_nicely("$fname: $!"); $must_close=1; }

    my $nrows = scalar @$seqs;
    my $ncols = length($$seqs[0]->{seq});

    print $fh "$nrows $ncols\n";

    for (my $i=0; $i<scalar @$seqs; $i++)
    {
        print $fh $seqs->[$i]->{seq};
        print $fh "\n";
    }

    if ( $must_close ) { close($fh); }
}


sub write_html_aln
{
    my ($seqs,$fname) = @_;

    my ($fh,$must_close);

    if ( ref($fname) eq 'GLOB' ) { $fh = $fname; }
    else { open($fh,">$fname") || exit_nicely("$fname: $!"); $must_close=1; }

    my $nrows = scalar @$seqs;
    my $ncols = length($$seqs[0]->{seq});

    print $fh "<table style='border-collapse:collapse'>\n";
    for (my $iseq=0; $iseq<scalar @$seqs; $iseq++)
    {
        print $fh "<tr><td>$$seqs[$iseq]->{id}$$seqs[$iseq]->{pad}";
        for (my $icol=0; $icol<$ncols; $icol++)
        {
            my $aa = substr($$seqs[$iseq]->{seq}, $icol,1);
            print $fh "<td class='aa_$aa'>$aa";
        }
        print $fh "\n";
    }
    print $fh "</table>\n";

    if ( $must_close ) { close($fh); }
}


sub write_histogram_aln
{
    my ($seqs,$fname,$options) = @_;

    if ( ref($fname) eq 'GLOB' ) { $fh = $fname; }
    else { open($fh,">$fname") || exit_nicely("$fname: $!"); $must_close=1; }

    my $fractions = ( $options && exists($$options{'fractions'}) && $$options{'fractions'} ) ? 1 : 0;
    my $nseqs = scalar @$seqs;

    my $len = length($$seqs[0]->{seq});
    for (my $icol=0; $icol<$len; $icol++)
    {
        my $histogram = {};
        for my $seq (@$seqs)
        {
            my $aa = substr($seq->{seq},$icol,1);
            $histogram->{$aa}++;
        }
        print $fh $icol+1;
        for my $aa (sort {$$histogram{$b}<=>$$histogram{$a}} keys %$histogram)
        {
            if ( $fractions )
            {
                printf $fh "\t$aa:%.2f", $$histogram{$aa}/$nseqs;
            }
            else
            {
                print $fh "\t$aa:$$histogram{$aa}";
            }
        }
        print $fh "\n";
    }

    if ( $must_close ) { close($fh); }
}

sub write_pir_aln
{
    my ($seqs,$fname) = @_;

    my ($fh,$must_close);

    if ( ref($fname) eq 'GLOB' ) { $fh = $fname; }
    else { open($fh,">$fname") || exit_nicely("$fname: $!"); $must_close=1; }

    for (my $i=0; $i<scalar @$seqs; $i++)
    {
        print $fh ">$seqs->[$i]->{id}\n";
        print $fh $seqs->[$i]->{seq};
        print $fh "\n";
    }

    if ( $must_close ) { close($fh); }
}


sub write_clustal_aln
{
    my ($seqs,$fname,$stats_line) = @_;

    my ($fh,$must_close);

    if ( ref($fname) eq 'GLOB' ) { $fh = $fname; }
    else { open($fh,">$fname") || exit_nicely("$fname: $!"); $must_close=1; }

    my @wrapped_seqs = ();
    my $ncols = 79 - length($$seqs[0]->{id}) - length($$seqs[0]->{pad});
    if ( $ncols <= 0 ) { exit_nicely("Eh, the sequences names too large? $ncols columns left.\n") }
    for (my $i=0; $i<scalar @$seqs; $i++)
    {
        push @wrapped_seqs, wrap($$seqs[$i]->{seq},$ncols);
    }
    print $fh "CLUSTAL\n\n";
    my $done=0;
    do
    {
        my $len;
        print $fh "\n";
        for (my $i=0; $i<scalar @$seqs; $i++)
        {
            my $part = shift(@{$wrapped_seqs[$i]});
            if ( !$part ) { $done=1; last; }
            my $out = "$seqs->[$i]->{id}$seqs->[$i]->{pad}$part";
            print $fh "$out\n";
            $len = length($out);
        }
        if ( !$done && $stats_line )
        {
            for (my $i=0; $i<$len; $i++)
            {
                print $fh ' ';
            }
            print $fh "\n";
        }
    }
    while (!$done);

    if ( $must_close ) { close($fh); }
}


sub wrap
{
    my ($string,$ncols) = @_;

    my @out = ();
    while ($string)
    {
        push @out, substr($string,0,$ncols,'');
    }
    return \@out;
}


sub read_table_aln
{
    my ($lines) = @_;

    my ($nrows,$ncols) = split(/\s+/,$$lines[0]);
    my @seqs = ();
    for (my $iline=1; $iline<$nrows; $iline++)
    {
        chomp($$lines[$iline]);
        push @seqs, { id=>$iline, ncols=>$ncols, seq=>$$lines[$iline] };
    }
    add_padding(\@seqs,10);
    return \@seqs;
}

sub read_clustal_aln
{
    my ($lines) = @_;

    my @seqs = ();
    my $i = 0;
    my ($indent,$ncolumns,$ncols);

    for (my $iline=1; $iline<scalar @$lines; $iline++)
    {
        my $line = $$lines[$iline];

        if ( $line=~/^\s*$/ ) { $i=0; next; }
        if ( $line=~/^(\S+)(\s+)(\S+)/ )
        {
            $ncols = length($3);
            if ( ! $indent )
            {
                $indent   = length($1) + length($2);
                $ncolumns = $ncols;
            }

            $seqs[$i]->{ncols} = $ncolumns;
            $seqs[$i]->{id}    = $1;
            $seqs[$i]->{pad}   = $2;
            $seqs[$i]->{seq}  .= $3;
            $i++;
            next; 
        }
        if ( $indent && ($line=~/^(\s{$indent})(.{$ncols})/) )
        {
            $seqs[$i]->{ncols} = $ncolumns;
            $seqs[$i]->{id}    = '';
            $seqs[$i]->{pad}   = $1;
            $seqs[$i]->{seq}  .= $2;
            $i++;
            next; 
        }
        exit_nicely("Could not parse line $iline: $line");
    }

    if ( $seqs[$#seqs]->{id} =~/^\s*$/ )
    {
        pop @seqs;      # The last sequence is only the clustal statistics
    }
    return \@seqs;
}


sub read_pir_aln
{
    my ($lines) = @_;

    my @seqs = ();
    my $ncolumns;

    my $id;
    my $iseq = -1;
    for (my $i=0; $i<scalar @$lines; $i++)
    {
        if ( $$lines[$i] =~ /^\s*$/ ) { next }
        if ( $$lines[$i] =~ /^>(.+)\s*$/  )
        {
            $iseq++;
            $id = $1;

            my $len = length($id);
            if ( !$maxlen || $len>$maxlen ) { $maxlen = $len }

            next;
        }
        if ( $id && $$lines[$i] =~ /^(\S+)\s*$/ )
        {
            $seqs[$iseq]->{id}   = $id;
            $seqs[$iseq]->{seq} .= $1;
            next;
        }
        exit_nicely("Could not parse line $i: $$lines[$i]");
    }
    add_padding(\@seqs);
    return \@seqs;
}


sub add_padding
{
    my ($seqs,$padding) = @_;

    my $max_pad;
    for (my $i=0; $i<scalar @$seqs; $i++)
    {
        my $len = length($$seqs[$i]->{id});
        if ( !$max_pad || $max_pad < $len ) { $max_pad = $len }
    }
    $max_pad++;

    if ( $padding && $max_pad > $padding )
    {
        exit_nicely("The request padding length is shorter than the longest sequence id: $padding < $max_pad.\n");
    }
    if ( !$padding ) { $padding = $max_pad }

    for (my $i=0; $i<scalar @$seqs; $i++)
    {
        my $len = length($$seqs[$i]->{id});
        my $pad = $padding-$len;

        $$seqs[$i]->{pad} = '';

        for (1..$pad)
        {
            $$seqs[$i]->{pad} .= ' ';
        }
    }
}



# Assuming that each sequence is on one (unwrapped) line.
sub read_oneline_aln
{
    my ($lines) = @_;

    my @seqs = ();
    my $ncolumns;

    for (my $i=0; $i<scalar @$lines; $i++)
    {
        if ( $$lines[$i] =~ /^(\S+)(\s+)(\S+)$/ )
        {
            $seqs[$i]->{id}  = $1;
            $seqs[$i]->{pad} = $2;
            $seqs[$i]->{seq} = $3;

            my $ncols = length($3);
            if ( $ncolumns && $ncols!=$ncolumns ) 
            {
                exit_nicely("Could not parse line $i. Different sequence length, expected $ncols, got $ncolumns.\n");
            }
            $ncolumns = $ncols;
            $seqs[$i]->{ncols} = $ncolumns;
            next;
        }
        if ( $$lines[$i]=~ /^\s*$/ ) { next; }
        exit_nicely("Could not parse line $i: \"$$lines[$i]\"");
    }
    return \@seqs;
}


# Order given as list 'id1,id2,...'. Only a subset may be given. In that case,
#   the listed sequences will be first (in the requested order) and the rest
#   will be left untouched.
#
sub reorder
{
    my ($seqs,$order) = @_;

    my @ordered_items  = split(/\s*,\s*/, $order);
    my $nordered_items = scalar @ordered_items;

    # The key 'order' will keep the information about the order.
    #   The first indices 0..($nordered_items-1) are reserved for
    #   the listed sequences, $nordered_items..(scalar @$seqs - 1) are
    #   used for keeping the unlisted sequences in place.
    #
    my $i = $nordered_items;
    for my $seq (@$seqs) { $$seq{'order'} = $i++; }

    $i = 0;
    for my $id (@ordered_items)
    {
        my $idxs = get_sequence_index($seqs,$id);
        if ( scalar @$idxs > 1 ) { Exit::exit_nicely("Multiple sequences match the ID $id.\n") }
        if ( !scalar @$idxs ) { Exit::exit_nicely("None of the sequences matches the ID $id.\n") }

        $$seqs[$$idxs[0]]->{'order'} = $i;
        $i++;
    }

    my @out = sort { $$a{'order'} <=> $$b{'order'} } @$seqs;
    return \@out;
}


sub get_sequence_index
{
    my ($seqs,$name,$options) = @_;

    my @ret = ();
    if ( !$name ) { return \@ret }

    my $exact = 0;
    if ( $options && exists($options->{exact}) && $options->{exact} ) { $exact=1; }

    for (my $i=0; $i<scalar @$seqs; $i++)
    {
        if ( $name eq $$seqs[$i]->{id} ) { push @ret, $i; }
    }
    if ( scalar @ret || $exact ) { return \@ret }

    for (my $i=0; $i<scalar @$seqs; $i++)
    {
        if ( $$seqs[$i]->{id} =~ /$name/i ) { push @ret,$i; }
    }

    return \@ret;
}


sub get_sequence
{
    my ($seqs,$name) = @_;

    my $idxs = get_sequence_index($seqs,$name);

    if ( ! scalar @$idxs ) { return '' }
    if ( scalar @$idxs != 1 ) { return 0; }

    return $$seqs[$$idxs[0]]->{seq};
}


sub get_position
{
    my ($seqs,$name,$pos) = @_;
    my $seq = get_sequence($seqs,$name);
    if ( ! $seq ) { return }
    return substr($seq,$pos,1);
}


# Returned indexes go from 0. Range may be given: ABC..XYZ.
sub find_substring
{
    my ($seq,$subseq) = @_;

    my @found = ();

    $subseq = uc($subseq);

    # First the simple case: We are looking for a substring which contains gaps (-),
    #   therefore we search the substring as it is.
    #
    if ( index($subseq,'-') != -1 )
    {
        my $i = -1;
        do
        {
            $i = index($seq,$subseq,$i+1);
            if ( $i==-1 ) { next }

            my @tmp = ();
            for (my $j=0; $j<length($subseq); $j++)
            {
                push @tmp,$i+$j;
            }
            push @found, \@tmp;
        }
        while ($i!=-1);
        return \@found;
    }

    # Now the more difficult case: The substring may be separated by gaps.
    #   Let's use regex's. It won't be very fast, but it is very easy to program.
    my @items = split(/\.\./,$subseq);
    my @patterns = ();
    for my $item (@items)
    {
        push @patterns, join('-*', split(//,$item));
    }
    my $pattern = join('.*', @patterns);
    my $offset = 0;
    while ( ($seq =~ /$pattern/) )
    {
        my $i = index($seq,$&);
        if ( $i==-1 ) { exit_nicely("FIXME: $subseq, $seq\n"); }
        
        my @tmp = ();
        for (my $j=0; $j<length($&); $j++)
        {
            push @tmp,$offset+$i+$j;
        }
        push @found, \@tmp;

        $offset += $i + length($&);
        $seq = $';
    }
    return \@found;
}


# Remove columns containing only gaps
sub remove_gap_columns
{
    my ($seqs) = @_;

    my $len = length($$seqs[0]{'seq'});

    my @gap_columns = ();
    for my $seq (@$seqs)
    {
        for (my $i=0; $i<$len; $i++)
        {
            my $aa = substr($$seq{'seq'}, $i, 1);
            if ( $aa ne '-' ) 
            { 
                $gap_columns[$i] = 0; 
            }
            elsif ( !exists($gap_columns[$i]) )
            {
                $gap_columns[$i] = 1; 
            }
        }
    }

    for my $seq (@$seqs)
    {
        for (my $i=$len-1; $i>=0; $i--)
        {
            if ( !$gap_columns[$i] ) { next }
            substr($$seq{'seq'},$i,1,'');
        }
    }
    return $seqs;
}


# Create substring of the alignment. The range may given either as a substring or as
#   column indexes starting from 1. Column indexes preceeded by a sequence ID are
#   interpreted as relative numbers (non-gapped positions), while indexes without
#   sequence ID are interpreted as absolute positions. When multiple strip definitions are
#   given, their order is not checked. Example: seq1:LKM-TMI,seq2:1-20,seq2:25,25
#
sub strip_alignment
{
    my ($seqs,$strip_string) = @_;

    my @ret = ();
    for my $seq (@$seqs)
    {
        push @ret, { id=>$seq->{id}, pad=>$seq->{pad}, seq=>'' };
    }

    my @strips = split(/,/, $strip_string);
    for my $strip (@strips)
    {
        my ($from,$to);
        if ( $strip=~/^(\S+):/ )
        {
            my $id = $1;
            my ($start,$end);

            if ( $strip=~/^\S+:(\S+)-(\S+)$/ )
            {
                $start = $1; 
                $end   = $2;
            }
            elsif ( $strip=~/^\S+:-(\S+)$/ )
            {
                $end   = $1;
            }
            elsif ( $strip=~/^\S+:(\S+)-$/ )
            {
                $start = $1;
            }
            elsif ( $strip=~/^\S+:(\S+)$/ )
            {
                $start = $1;
                $end   = $1;
            }

            my $seq = Alignments::get_sequence($seqs,$id);
            if ( !$seq ) { Exit::exit_nicely("No such sequence \"$id\" in the alignment.\n") }
            my $gapped = $seq;

            $seq =~ s/-//g;
            my ($start_pos,$end_pos);
            if ( $start && $start=~/\d+/ )
            {
                $start_pos = $start - 1;
            }
            elsif ( !$start )
            {
                $start_pos = 0;
            }
            else
            {
                $start_pos = index($seq,$start);
                if ( $start_pos<0 ) { Exit::exit_nicely("The string \"$start\" was not found in $id.\n"); }
                if ( index($seq,$start,$start_pos+1)!=-1 ) { Exit::exit_nicely("The substring \"$start\" is not unique.\n") }
            }
            if ( $end && $end=~/\d+/ )
            {
                $end_pos = $end - 1;
            }
            elsif ( !$end )
            {
                $end_pos = length($seq) - 1;
            }
            else
            {
                $end_pos = index($seq,$end);
                if ( $end_pos<0 ) { Exit::exit_nicely("The string \"$end\" was not found in $id.\n"); }
                if ( index($seq,$end,$end_pos+1)!=-1 ) { Exit::exit_nicely("The substring \"$end\" is not unique.\n") }
                $end_pos += length($end) - 1;
            }

            my $rel_to_abs = Alignments::relative_to_absolute($gapped,$start_pos,$end_pos);
            $from = $$rel_to_abs[0];
            $to   = $$rel_to_abs[1];
        }
        elsif ( $strip=~/^(\d+)-(\d+)$/ )
        {
            $from = $1-1;
            $to   = $2-1;
        }
        elsif ( $strip=~/^-(\d+)$/ )
        {
            $from = 0;
            $to   = $1-1;
        }
        elsif ( $strip=~/^(\d+)-$/ )
        {
            $from = $1-1;
            $to   = length($$seqs[0]->{seq})-1;
        }
        elsif ( $strip=~/^-(\d+)-$/ )   # Count index from the end
        {
            my $len = length($$seqs[0]->{seq});
            $from = $len - $1;
            $to   = $len - 1;
        }
        elsif ( $strip=~/^(\d+)$/ )
        {
            $from = $1-1;
            $to   = $1-1;
        }
        else
        {
            Exit::exit_nicely("Could not parse the string \"$strip\".\n"); 
        }

        for (my $i=0; $i<scalar @$seqs; $i++)
        {
            my $abs_to_rel = absolute_to_relative($$seqs[$i]->{seq});
            $ret[$i]->{from} = $$abs_to_rel[$from];
            $ret[$i]->{to}   = $$abs_to_rel[$to];

            my $tmp = $$seqs[$i]->{seq};
            $tmp = substr($tmp,0,$to + 1);

            if ( $from>=length($tmp) ) { Exit::exit_nicely("FIXME: $from - $to, ",length($$seqs[$i]->{seq})," $$seqs[$i]->{seq}\n") }
            $tmp = substr($tmp,$from);

            $ret[$i]->{seq} .= $tmp;
        }
    }
    return \@ret;
}


# The first alignment will be modified.
sub glue_alignments
{
    my ($aln1,$aln2) = @_;

    if ( !$aln1 ) { return $aln2 }
    if ( !$aln2 ) { return $aln1 }

    if ( scalar @$aln1 != @$aln2 ) { Exit::exit_nicely("The alignments have different number of rows (",scalar @$aln1,", ", scalar @$aln2,").\n") }

    for (my $iseq=0; $iseq<scalar @$aln1; $iseq++)
    {
        $$aln1[$iseq]->{seq} .= ' ' . $$aln2[$iseq]->{seq};
    }
    return $aln1;
}


# Like perl's substr, but applied on all sequences.
sub substring
{
    my ($seqs,$offset,$length,$replacement) = @_;

    for my $seq (@$seqs)
    {
        if ( $length && $replacement )
        {
            substr($seq->{seq},$offset,$length,$replacement);
        }
        elsif ( $length )
        {
            $seq->{seq} = substr($seq->{seq},$offset,$length);
        }
        else
        {
            $seq->{seq} = substr($seq->{seq},$offset);
        }
    }
}




# Relative = positions without gaps, absolute = including gaps. Indexes must go from 0.
# If the second argument is not given, all positions will be given.
#
sub relative_to_absolute
{
    my ($sequence,@positions) = @_;

    my $relative_idx = 0;
    my $relative_to_absolute = {};
    my $len = length($sequence);
    for (my $absolute_idx=0; $absolute_idx<$len; $absolute_idx++)
    {
        my $letter = substr($sequence,$absolute_idx,1);
        
        if ( $letter eq  '-' ) { next; }

        $relative_to_absolute->{$relative_idx} = $absolute_idx;
        $relative_idx++;
    }

    my @abs_positions = ();
    if ( ! scalar @positions )
    {
        for my $relative_idx (sort {$a<=>$b} keys %$relative_to_absolute)
        {
            push @abs_positions, $relative_to_absolute->{$relative_idx};
        }
        return \@abs_positions;
    }

    for my $position (@positions)
    {
        if ( ref($position) eq 'ARRAY' )
        {
            for my $relative_idx (@$position)
            {
                push @abs_positions, exists($relative_to_absolute->{$relative_idx}) ? $relative_to_absolute->{$relative_idx} : -1;
            }
        }
        else
        {
            push @abs_positions, exists($relative_to_absolute->{$position}) ? $relative_to_absolute->{$position} : -1;
        }
    }

    return \@abs_positions;
}


# The same as relative_to_absolute_positions, but in the opposite direction.
#
sub absolute_to_relative
{
    my ($sequence,@positions) = @_;

    my $relative_idx = 0;
    my $absolute_to_relative = {};
    my $len = length($sequence);
    for (my $absolute_idx=0; $absolute_idx<$len; $absolute_idx++)
    {
        my $letter = substr($sequence,$absolute_idx,1);
        if ( $letter eq '-' )
        {
            $absolute_to_relative->{$absolute_idx} = -1;
            next;
        }
        $absolute_to_relative->{$absolute_idx} = $relative_idx;
        $relative_idx++;
    }
    
    my @rel_positions = ();
    if ( !scalar @positions )
    { 
        for (my $absolute_idx=0; $absolute_idx<$len; $absolute_idx++)
        {
            push @rel_positions, $absolute_to_relative->{$absolute_idx}; 
        }
        return \@rel_positions;
    }

    for my $position (@positions)
    {
        if ( ref($position) eq 'ARRAY' )
        {
            for my $absolute_idx (@$position)
            {
                push @rel_positions, exists($absolute_to_relative->{$absolute_idx}) ? $absolute_to_relative->{$absolute_idx} : -1;
            }
        }
        else
        {
            push @rel_positions, exists($absolute_to_relative->{$position}) ? $absolute_to_relative->{$position} : -1;
        }
    }

    return \@rel_positions;
}


sub relative_to_relative
{
    my ($seq_a,$seq_b,@positions) = @_;

    my $abs_positions = relative_to_absolute($seq_a,@positions);
    return absolute_to_relative($seq_b,$abs_positions);
}


# File names can be given instead of alignments.
#
sub aln_absolute_to_relative
{
    my ($id1,$aln1,$id2,$aln2) = @_;

    if ( ref $aln1 ne 'ARRAY' ) { $aln1 = parse_alignment($aln1) }
    if ( ref $aln2 ne 'ARRAY' ) { $aln2 = parse_alignment($aln2) }

    my $seq1 = get_sequence($aln1,$id1);
    my $seq2 = get_sequence($aln2,$id2);

    my ($aln,$fname) = run_clustal( [ { 'id'=>'id1', 'seq'=>$seq1}, { 'id'=>'id2', 'seq'=>$seq2}], { 'delete'=>1 });

    my $abs_to_rel = absolute_to_relative($seq1);
    my $rel_to_rel = relative_to_relative(get_sequence($aln,'id1'),get_sequence($aln,'id2'),@$abs_to_rel);

    return $rel_to_rel;

    #   for (my $i=0; $i<scalar @$abs_to_rel; $i++)
    #   {
    #       my $j = $$rel_to_rel[$i];
    #       my $a = substr($seq1,$i,1);
    #
    #       if ( $a eq '-' ) { print STDERR "$i .. $a -\n"; next }
    #       my $b = substr($seq2,$j,1);
    #
    #       print STDERR "$i .. $a $b\n";
    #   }
}


sub aln_relative_to_relative
{
    my ($id1,$aln1,$id2,$aln2) = @_;

    if ( ref $aln1 ne 'ARRAY' ) { $aln1 = parse_alignment($aln1) }
    if ( ref $aln2 ne 'ARRAY' ) { $aln2 = parse_alignment($aln2) }

    my $seq1 = get_sequence($aln1,$id1);
    my $seq2 = get_sequence($aln2,$id2);

    my ($aln,$fname) = run_clustal( [ { 'id'=>'id1', 'seq'=>$seq1}, { 'id'=>'id2', 'seq'=>$seq2}], { 'delete'=>1 });

    my $rel_to_rel = relative_to_relative(get_sequence($aln,'id1'),get_sequence($aln,'id2'));

    return $rel_to_rel;
}


sub run_clustal
{
    my ($seqs,$options) = @_;

    if ( !scalar @$seqs ) { exit_nicely("Passing an empty array to run_clustal?\n") }

    my $delete   = 0;
    my $tmp_file = '/tmp/Alignments-'. $$;
    if ( $options )
    {
        if ( exists($$options{'delete'}) && $$options{'delete'}==1 ) { $delete=1; }
        if ( exists($$options{'prefix'}) ) { $tmp_file=$$options{'prefix'}; }
    }

    open(FILE,">$tmp_file.inp") || exit_nicely("$tmp_file.inp: $!");
    for my $seq (@$seqs)
    {
        print FILE ">$$seq{id}\n$$seq{seq}\n";
    }
    close(FILE);

    `clustalw $tmp_file.inp`;
    my $aln = parse_alignment("$tmp_file.aln");
    if ( ! scalar @$aln ) { exit_nicely("Could not read the result $tmp_file.aln, empty alignment??\n") }

    unlink("$tmp_file.inp");
    unlink("$tmp_file.dnd");
    if ( $delete ) { `rm -f $tmp_file*`; }

    return ($aln,"$tmp_file.aln");
}



# The subroutine expects on input a hash with aminoacids abundances, e.g. $letters->{A} = 5, $letters->{B} = 1;
#   It returns references to two arrays: scaled relative entropies and average E1-5 values.
#
# The code was modified to tackle any number of bins and properties.
#
sub calc_entropy
{
    my ($letters) = @_;

    if ( !defined($total_entropy_approach) ) { exit_nicely("Have you forgot to call Alignments::select_total_entropy_method() from your script?\n"); }

    my $n_no_gaps = 0;
    my $n_with_gaps = 0;
    for my $aa (keys %$letters)
    {
        $n_with_gaps += $letters->{$aa};
        if ( $aa eq '-' ) { next }
        $n_no_gaps += $letters->{$aa};
    }

    my @entropies = ();
    my @prop_avgs = ();
    my @prop_devs = ();
    my $total = 0;
    my $similarity = 0;
    my $pcdist = 0;     # physicochemical distance

    if ( !$n_no_gaps ) 
    { 
        for my $i (0..$Vectorlib::nprops)
        {
            $entropies[$i] = 0;
            $prop_avgs[$i] = 0;
            $prop_devs[$i] = -1;
        }
        return (\@entropies,\@prop_avgs,\@prop_devs,$total,$similarity,$pcdist); 
    }

    # First calculate the average physicochemical distance/similarity - for 
    #   conserved columns is the calculation very fast, as the identical letters 
    #   are grouped together.
    #
    # NOTE: if you change this, update also documentation in WebPCPMer:web/Help/mmaker_motif_types.inc.
    #
    my @aas = keys %$letters;
    for (my $i=0; $i<scalar @aas; $i++)
    {
        if ( $aas[$i] eq '-' ) { next; }

        if ( !exists($$Vectorlib::vectors{1}{$aas[$i]}) ) 
        { 
            Exit::exit_nicely(
                    "Unknown descriptor value E1 for amino acid \"$aas[$i]\"\n",
                    "This usually means that your multiple alignment contains non-standard aminoacids.\n",
                    "Either edit the alignment manually, or use the substitution feature.\n"
                    ); 
        }

        for (my $j=0; $j<$i; $j++)
        {
            if ( $aas[$j] eq '-' ) { next; }
    
            my $dist = 0;
            for my $iprop (1..$Vectorlib::nprops)
            {
                $dist += ($$Vectorlib::vectors{$iprop}{$aas[$i]} - $$Vectorlib::vectors{$iprop}{$aas[$j]})**2;
            }
            $pcdist += $$letters{$aas[$i]} * $$letters{$aas[$j]} * sqrt($dist);
        }
    }
    if ( $n_no_gaps>1 )
    {
        $pcdist /= $n_no_gaps*($n_no_gaps-1)*0.5;
        $similarity = exp(-$pcdist*0.1) * $n_no_gaps/$n_with_gaps;
    }
    elsif ( $n_with_gaps==1 ) { $similarity=1; $pcdist=0; }
    else { $similarity=0; $pcdist=-1;}


    # Now calculate the entropies, mean values and deviations
    for my $property (1..$Vectorlib::nprops)
    {
        my $intervals = $Vectorlib::intervals->{$property};
        my $natfreqs  = $Vectorlib::natfreqs->{$property};
        my $vectors   = $Vectorlib::vectors->{$property};
        my @probability_histogram = ();
        for my $i (0..scalar @$intervals-1) { $probability_histogram[$i] = 0 }

        # Calculate the avg property value and fill bins (i.e. create probability
        #   distribution).
        for my $aa (keys %$letters)
        {
            if ( $aa eq '-' ) { next; }
            if ( !exists($vectors->{$aa}) ) 
            { 
                Exit::exit_nicely(
                        "Unknown value E$property for amino acid \"$aa\"\n",
                        "This usually means that your multiple alignment contains non-standard aminoacids.\n",
                        "Either edit the alignment manually, or use the substitution feature.\n"
                        ); 
            }

            my $property_value = $vectors->{$aa};
            my $count = $letters->{$aa};

            $prop_avgs[$property-1] += $vectors->{$aa} * $letters->{$aa} / $n_no_gaps;

            for (my $i=0; $i<scalar @$intervals-1; $i++)
            {
                if ( $property_value > $$intervals[$i] && $property_value <= $$intervals[$i+1] )
                {
                    $probability_histogram[$i] += $count;
                    last;
                }
            }
        }
        # Calculate the property's deviation as sqrt[(Xi-mean(X))**2/N]
        $prop_devs[$property-1] = 0;
        for my $aa (keys %$letters)
        {
            $prop_devs[$property-1] += ($vectors->{$aa} - $prop_avgs[$property-1])**2  * $letters->{$aa} / $n_no_gaps;
        }
        $prop_devs[$property-1] = sqrt($prop_devs[$property-1]);

        # This is not nice: The absolutely conserved positions are recognised in motifmaker.pm:makeprofile() by
        #   the test $dev>1e-10. If there are some gaps and $dev<1e-10, set $dev=1e-9.
        #
        if ( $n_with_gaps!=$n_no_gaps && $prop_devs[$property-1]<1e-10 ) { $prop_devs[$property-1]=1e-9; }

        # Now probability_histogram[i] is equal to the probability that an aminoacid from the given set of aminoacids
        #   falls in the i-th bin. If the set of aminoacids contains just one letter, only one probability_histogram[i]
        #   will have non-zero value and it would be equal to 1.

        my $entropy = 0;
        for my $bin (0..$Vectorlib::nbins-1)
        {
            # Replace gaps by average noise
            $probability_histogram[$bin] += $$natfreqs[$bin] * ($n_with_gaps-$n_no_gaps);
            $probability_histogram[$bin] /= $n_with_gaps;

            if ( !$probability_histogram[$bin] ) { next; }
            $entropy += $probability_histogram[$bin] * log($probability_histogram[$bin]/$$natfreqs[$bin]) / log(2);
        }

        $entropy *= $scaleS[$property-1];
        push @entropies,$entropy;

        if ( $total_entropy_approach eq 'unscaled_max' || $total_entropy_approach eq 'ori' )
        {
            if ( !$total || $total<$entropy ) { $total=$entropy }
        }
        elsif ( $total_entropy_approach eq 'unscaled_sum' || $total_entropy_approach eq 'scaled_sum' )
        {
            $total += $entropy;
        }
        else { Exit::exit_nicely("Unknown total entropy approach '$total_entropy_approach'.\n"); }
    }

    $total /= $entropy_limits->{largest};

    return (\@entropies,\@prop_avgs,\@prop_devs,$total,$similarity,$pcdist);
}




# Go through the 20 aminoacids and decide which lies closest to
#   the given point in the five-dimensional property space.
#   For the notes about scaling of E1-E5 space, see create_consensus_sequence.
#   If a hash $subset is given, only the aminoacids from the subset will be
#   selected from. 
#
# NOTE: The variable $e5point points to an array with five values
#   and it must be indexed starting from 1.
#
sub find_closest_amino_acid
{
    my ($e5point,$e5scales,$options) = @_;

    my $subset  = 0;
    my $verbose = 0;
    if ( $options )
    {
        if ( exists($$options{'subset'}) && $$options{'subset'} ) { $subset = $$options{'subset'} }
        if ( exists($$options{'verbose'}) ) { $verbose = $$options{'verbose'} }
    }

    my $min_sq_dist = -1;
    my $min_aa = '-';
    for my $aa ('A','R','N','D','C','Q','E','G','H','I','L','K','M','F','P','S','T','W','Y','V')
    {
        if ( $subset && !exists($subset->{$aa}) ) { next; }

        my $sq_dist = 0;
        for my $e (1..$Vectorlib::nprops)
        {
            $sq_dist += (($Vectorlib::vectors->{$e}->{$aa} - $$e5point[$e-1]) * $$e5scales[$e-1]) **2;
        }
        if ( $verbose )
        {
            printf "\t$aa \t %.2f\n", $sq_dist;
        }
        if ( $min_sq_dist<0 || $sq_dist<$min_sq_dist ) 
        {
            $min_sq_dist = $sq_dist;
            $min_aa = $aa;
        }
    }
    return $min_aa;
}



# For every position of the multiple sequence alignment five average property 
#   values are calculated and then the amino acid with the least Euclidean 
#   distance from this point in the five-dimensional property space will be
#   selected. The five-dimensional property space will be scaled (biased) for each
#   position so that those amino acids will be taken preferentially, which
#   have similar values for the most significant property. For example, if a
#   given position has high entropy in E1 and small entropies in E2-E5, the space
#   will be scaled so that the E1 distances between aminoacids will become more
#   relevant than the distances in E2-E5. This feature may be turned off by passing
#   a hash containing the key nobias=>1 as a parameter.
#
# For each nongapped position, a scaled sum of relative entropies will be returned, so that
#   the variability of the consensus may be checked. The values are from the range
#   0-1. 100% conserved Leu has entropy equal to $least_abs_conserved_relsum = $least_abs_conserved_abssum/$max_possible_entropies_sum.
#
# The parameters least_entrps and least_smls is useful when creating "grand alignments", that is,
#   an alignment is used to create consensus sequence and a "grand consensus" out of the
#   consensus' alignment is generated. See the description in grand-aln.
#
# NOTE: there is still one problem: Consider a column T:16 S:5 divided into two groups
#   T:13 and S:5 T:3 with the alignment of consensus sequences S:1 and T:1. The total
#   consensus will be S instead of T! (Example from mosquito-tick E protein aln, S298.
#
sub create_consensus_sequence
{ 
    my ($aligned_seqs,$options,$other) = @_;

    if ( $other ) { Exit::exit_nicely("The create_consensus_sequence interface has changed!!\n") }

    my $nobias   = 0;
    my $restrict = 1;
    my $least_entrps = 0;
    my $least_smls   = 0;  # smls stands for similarities

    if ( $options )
    {
        if ( ref($options) ne 'HASH' ) { Exit::exit_nicely("The create_consensus_sequence interface has changed!!\n") }

        if ( exists($$options{'nolimits'}) && $$options{'nolimits'}==1 ) { $restrict = 0; }
        if ( exists($$options{'nobias'}) && $$options{'nobias'}==1 ) { $nobias = 1; }
        if ( exists($$options{'least_entrps'}) ) { $least_entrps = $$options{'least_entrps'}; }
        if ( exists($$options{'least_smls'}) ) { $least_smls = $$options{'least_smls'}; }
    }

    my @consensus    = ();
    my @entropies    = ();
    my @similarities = ();

    my $nseqs = scalar @$aligned_seqs;
    my $len   = length($$aligned_seqs[0]->{seq});
    for (my $ipos=0; $ipos<$len; $ipos++)
    {
        my $letters = {};
        for my $seq (@$aligned_seqs)
        {
            my $letter = substr($seq->{seq},$ipos,1);
            $letters->{$letter}++;
        }

        my ($entrps,$avgs,$devs,$total,$similarity) = calc_entropy($letters);
        my $restrict_set = $restrict ? $letters : 0;
        my $space_bias;
        if ( $nobias )
        {
            for my $i (0..$Vectorlib::nprops) { push @$space_bias,1; }
        }
        else 
        { 
            $space_bias = $entrps;  # This works well, because entropies are from the range <0,1>.
        }
        my $closest = find_closest_amino_acid($avgs, $space_bias, { 'subset'=>$restrict_set });

        if ( $least_entrps && scalar @$least_entrps && $total>$$least_entrps[$ipos] ) { $total = $$least_entrps[$ipos] }
        if ( $least_smls && scalar @$least_smls && $similarity>$$least_smls[$ipos] ) { $similarity = $$least_smls[$ipos] }
        if ( $total == -1 ) { $closest = '-' }

        push @consensus,$closest;
        push @entropies,$total;
        push @similarities,$similarity;
    }
    return (\@consensus,\@entropies,\@similarities);
}


sub create_gnd_consensus
{
    require Flavitrack;

    my ($aln_file) = @_;
    my $virus_alns = {};

    my $aln = parse_alignment($aln_file,{'substitute'=>$common_substitutions});
    for my $seq (@$aln)
    {
        my $lp = Flavitrack::parse_license_plate($seq->{id});
        my $virus = $lp->{virus};
        
        push @{$virus_alns->{$virus}}, $seq;
    }

    print STDERR scalar keys %$virus_alns, " groups for $aln_file...\n";

    my $cons;
    my @grand_aln    = ();
    my @least_entrps = ();
    my $entrps = \@least_entrps;
    for my $virus (keys %$virus_alns)
    {
        ($cons,$entrps) = create_consensus_sequence($virus_alns->{$virus}, $entrps);

        my $seq;
        $seq->{seq} = join('',@$cons);
        $seq->{id}  = $virus;

        push @grand_aln, $seq;
    }

    ($cons,$entrps) = create_consensus_sequence(\@grand_aln, $entrps);
    return ($cons,$entrps);
}


sub parse_alignment_entropies
{
    my ($aln_file) = @_;

    my $entrps = {};

    my $entrps_file = ( $aln_file=~/\.aln$/i ) ? "$`.entrps" : "$aln_file.entrps";
    if ( ! -e $entrps_file ) { return $entrps }

    open(ENTRPS,"<$entrps_file") || exit_nicely("$entrps_file: $!");
    while (my $line=<ENTRPS>)
    {
        chomp($line);
        my @items = split(/\t/,$line);
        my $seq_name = shift(@items);
        $entrps->{$seq_name} = \@items;
    }
    close(ENTRPS);

    return $entrps;
}

sub parse_alignment_similarities
{
    my ($aln_file) = @_;

    my $smls = {};

    my $smls_file = ( $aln_file=~/\.aln$/i ) ? "$`.smls" : "$aln_file.smls";
    if ( ! -e $smls_file ) { return $smls }

    open(FILE,"<$smls_file") || exit_nicely("$smls_file: $!");
    while (my $line=<FILE>)
    {
        chomp($line);
        my @items = split(/\t/,$line);
        my $seq_name = shift(@items);
        $smls->{$seq_name} = \@items;
    }
    close(FILE);

    return $smls;
}

# This is used when working with multiple organisms with high subtype
# variability: If we want to analyze motifs common, for example, to all
# flaviviruses, we cannot work with all sequences available. Having many more DV2
# sequences (hundreds of them) than AHF sequences (only one), the statistics
# would be biased toward DV2. If we selected randomly only one representative
# from DV2, the result could be also misleading. Therefore, we create consensus
# sequence for each group (see the programs grand-aln and
# create-consensus-sequence) and look for motifs in this "grand-alignment".
# However, there is a problem: Consider a very variable position. Consensus at
# this position may be accidently identical for both groups, but in the
# grand-alignment this position would appear as a significant and absolutely
# conserved residue. To prevent this, the entropies for each position are
# remembered and the minimum of the newly calculated and remembered entropy is
# used.
# The remembered entropys is calculated as an average of entropies of subalignments.
#
sub get_least_alignment_entropies
{
    my ($aln_file,$seqs,$entrps) = @_;

    my @leasts = ();
    if ( !$entrps )
    {
        $entrps = parse_alignment_entropies($aln_file);
    }
    if ( scalar keys %$entrps == 0 ) 
    { 
        return \@leasts;
    }

    my $nseqs = scalar @$seqs;
    my $len   = length($$seqs[0]->{seq});
    for (my $i=0; $i<$len; $i++)
    {
        my $avg = 0;
        for my $seq (@$seqs)
        {
            my $entropy = $entrps->{$seq->{id}}->[$i];
            $avg += $entropy>0 ? $entropy : 0;
        }
        push @leasts, $avg/$nseqs;
    }
    return \@leasts;
}


# Same as get_least_alignment_entropies, but for similarities.
sub get_least_alignment_similarities
{
    my ($aln_file,$seqs,$smls) = @_;

    my @leasts = ();
    if ( !$smls )
    {
        $smls = parse_alignment_similarities($aln_file);
    }
    if ( scalar keys %$smls == 0 ) 
    { 
        return \@leasts;
    }

    my $nseqs = scalar @$seqs;
    my $len   = length($$seqs[0]->{seq});
    for (my $i=0; $i<$len; $i++)
    {
        my $avg = 0;
        for my $seq (@$seqs)
        {
            $avg += $smls->{$seq->{id}}->[$i];
        }
        push @leasts, $avg/$nseqs;
    }
    return \@leasts;
}


1;

