#!/usr/bin/perl -w
#
# PDB specification: 
#   http://bmerc-www.bu.edu/needle-doc/latest/atom-format.html
#
#   http://www.biochem.ucl.ac.uk/~roman/procheck/manual/manappb.html
#   ------------------------------------------------------------------------------------------
#   Field |    Column    |    Column    | FORTRAN |                                         
#     No. |     range    |     range    | format  | Description                                   
#   ------------------------------------------------------------------------------------------
#      1. |    1 -  6    |    0 -  5    |   A6    | Record ID (eg ATOM, HETATM)       
#      2. |    7 - 11    |    6 - 10    |   I5    | Atom serial number                            
#      -  |   12 - 12    |   11 - 11    |   1X    | Blank                                         
#      3. |   13 - 16    |   12 - 15    |   A4    | Atom name (eg " CA " , " ND1")   
#      4. |   17 - 17    |   16 - 16    |   A1    | Alternative location code (if any)            
#      5. |   18 - 20    |   17 - 19    |   A3    | Standard 3-letter amino acid code for residue 
#      -  |   21 - 21    |   20 - 20    |   1X    | Blank                                         
#      6. |   22 - 22    |   21 - 21    |   A1    | Chain identifier code                         
#      7. |   23 - 26    |   22 - 25    |   I4    | Residue sequence number                       
#      8. |   27 - 27    |   26 - 26    |   A1    | Insertion code (if any)                       
#      -  |   28 - 30    |   27 - 29    |   3X    | Blank                                         
#      9. |   31 - 38    |   30 - 37    |  F8.3   | Atom's x-coordinate                         
#     10. |   39 - 46    |   38 - 45    |  F8.3   | Atom's y-coordinate                         
#     11. |   47 - 54    |   46 - 53    |  F8.3   | Atom's z-coordinate                         
#     12. |   55 - 60    |   54 - 59    |  F6.2   | Occupancy value for atom                      
#     13. |   61 - 66    |   60 - 65    |  F6.2   | B-value (thermal factor)                    
#      -  |   67 - 67    |   66 - 66    |   1X    | Blank                                         
#     14. |   68 - 70    |   67 - 69    |   I3    | Footnote number                               
#   ------------------------------------------------------------------------------------------
#

package PDB;

use Exit qw(exit_nicely);

$three_to_one =
{
    'ala'  =>   'a',
    'arg'  =>   'r',
    'asn'  =>   'n',
    'asp'  =>   'd',
    'cys'  =>   'c',
    'glu'  =>   'e',
    'gln'  =>   'q',
    'gly'  =>   'g',
    'his'  =>   'h',
    'ile'  =>   'i',
    'leu'  =>   'l',
    'lys'  =>   'k',
    'met'  =>   'm',
    'phe'  =>   'f',
    'pro'  =>   'p',
    'ser'  =>   's',
    'thr'  =>   't',
    'trp'  =>   'w',
    'tyr'  =>   'y',
    'val'  =>   'v',
};


for my $three (keys %$three_to_one)
{
    $one_to_three->{$three_to_one->{$three}}     = $three;
    $one_to_three->{uc($three_to_one->{$three})} = uc($three);
    $three_to_one->{uc($three)} = uc($three_to_one->{$three});
}


# The PDB is returned as a list (array) of residues. Each residue
#   is represented as a hash containing the fields named as record_id,
#   atom_id, etc. (see the push call below).
#
sub read_pdb
{
    my ($fname) = @_;

    my ($fh,$must_close);
    
    if ( ref($fname) eq 'GLOB' ) { $fh = $fname; }
    else { open($fh,"<$fname") || exit_nicely("$fname: $!"); $must_close=1; }

    my $residues = {};
    my @pdb = ();
    my ($prev_residue,$dupl_residue);
    while (my $line=<$fh>)
    {
        # ATOM      1  HN2 TYR     1      -0.338   0.478   0.828
        if ( !($line=~/^ATOM/) ) { next; }

        my $record_id    = substr($line,0,6);
        my $atom_number  = substr($line,6,5);
        my $atom_name    = substr($line,12,4);
        my $loc_code     = substr($line,16,1);
        my $res_name     = substr($line,17,3);
        my $chain_id     = substr($line,21,1);
        my $res_number   = substr($line,22,5);
        my $insert_code  = substr($line,26,1);
        my $x            = substr($line,30,8);
        my $y            = substr($line,38,8);
        my $z            = substr($line,46,8);
        my $ocupancy_val = substr($line,54,6);
        my $b_value      = substr($line,60,6);
        my $footnote     = length($line)>=67+3 ? substr($line,67,3) : '';

        $res_number =~ s/\s+//g; 
        my $residue_id = $chain_id . '/' . lc($res_name) . $res_number;   # E.g. lys236
        
        if ( $prev_residue && $prev_residue ne $res_number )
        {
            $dupl_residue = 0;
            if ( exists($residues->{$residue_id}) )
            {
                $dupl_residue = 1;
                do
                {
                    $dupl_residue++;
                }
                while ( exists($residues->{$residue_id .'_'. $dupl_residue}) )
            }
        }
        if ( $dupl_residue ) { $residue_id .= '_' . $dupl_residue; }
        $prev_residue = $res_number;

        $x =~ s/\s+//g; 
        $y =~ s/\s+//g; 
        $z =~ s/\s+//g; 

        push @{$residues->{$residue_id}}, 
        { 
            record_id     => $record_id,
            atom_id       => $atom_number, 
            name          => $atom_name,
            loc_code      => $loc_code,
            res_name      => $res_name,
            chain_id      => $chain_id,
            res_number    => $res_number,
            insert_code   => $insert_code,
            x             => $x,
            y             => $y,
            z             => $z,
            ocupancy_val  => $ocupancy_val,
            b_value       => $b_value,
            footnote      => $footnote
        };

        my $uc_res_name = uc($res_name);
        my $one_letter_code = exists($$three_to_one{$uc_res_name}) ? $$three_to_one{$uc_res_name} : '';

        if ( !scalar @pdb || $pdb[$#pdb]->{res_id} ne $residue_id ) 
        { 
            push @pdb,{ res_id=>$residue_id, res_number=>$res_number, chain=>$chain_id, name=>$res_name, one_letter=>$one_letter_code };
        }
    }
    
    if ( $must_close ) { close($fh); }

    for (my $i=0; $i<scalar @pdb; $i++)
    {
        my $id = $pdb[$i]->{res_id};
        $pdb[$i]->{atoms} = $residues->{$id};
    }

    return \@pdb;
}


sub write_pdb
{
    my ($pdb,$fname,$options) = @_;

    my ($fh,$must_close);
    
    if ( ref($fname) eq 'GLOB' ) { $fh = $fname; }
    else { open($fh,">$fname") || exit_nicely("$fname: $!"); $must_close=1; }

    my $renumber = 0;
    if ( $options && exists($$options{'renumber'}) && $$options{'renumber'}==1 ) { $renumber=1; }

    my ($ires,$iatom);
    for (my $ires=0; $ires<scalar @$pdb; $ires++)
    {
        $res = $$pdb[$ires];

        my $atoms = $res->{atoms};
        for (my $iatom=0; $iatom<scalar @$atoms; $iatom++)
        {
            $atom = $$atoms[$iatom];

            printf $fh "%6s%5d %4s%1s%3s %1s%4d%1s   %8.3f%8.3f%8.3f%6.2f%6.2f %s\n",
                $atom->{record_id}, 
                $renumber ? $iatom+1 : $atom->{atom_id},
                $atom->{name},
                $atom->{loc_code},
                $atom->{res_name},
                $atom->{chain_id},
                $renumber ? $ires+1 : $atom->{res_number},
                $atom->{insert_code},
                $atom->{x},
                $atom->{y},
                $atom->{z},
                $atom->{ocupancy_val},
                $atom->{b_value},
                $atom->{footnote};
        }
    }

    if ( $must_close ) { close($fh); }
}



sub read_pdb_residues
{
    my ($fname) = @_;

    my ($fh,$must_close);
    
    if ( ref($fname) eq 'GLOB' ) { $fh = $fname; }
    else { open($fh,"<$fname") || exit_nicely("$fname: $!"); $must_close=1; }

    my $residues = {};
    while (my $line=<$fh>)
    {
        # ATOM      1  HN2 TYR     1      -0.338   0.478   0.828
        if ( !($line=~/^ATOM/) ) { next; }

        my $record_id    = substr($line,0,6);
        my $atom_number  = substr($line,6,5);
        my $atom_name    = substr($line,12,4);
        my $loc_code     = substr($line,16,1);
        my $name         = substr($line,17,3);
        my $chain_id     = substr($line,21,1);
        my $seq_number   = substr($line,22,5);
        my $insert_code  = substr($line,26,1);
        my $x            = substr($line,30,8);
        my $y            = substr($line,38,8);
        my $z            = substr($line,46,8);
        my $ocupancy_val = substr($line,54,6);
        my $b_value      = substr($line,60,6);
        my $footnote     = substr($line,67,3);

        $seq_number =~ s/\s+//g; 
        my $id = $chain_id . '/' . lc($name) . $seq_number;   # E.g. lys236
        
        $x =~ s/\s+//g; 
        $y =~ s/\s+//g; 
        $z =~ s/\s+//g; 

        my $atom = { x=>$x, y=>$y, z=>$z };
        push @{$residues->{$id}}, $atom;
    }
    
    if ( $must_close ) { close($fh); }

    return $residues;
}


sub get_sequence
{
    my ($pdb,$options) = @_;

    my $output_array = 0;
    my $upper_case   = 0;
    my $chain;
    if ( $options )
    {
        if ( exists($$options{'array'}) && $$options{'array'}==1 ) { $output_array = 1; }
        if ( exists($$options{'uc'}) && $$options{'uc'}==1 ) { $upper_case = 1; }
        if ( exists($$options{'chain'}) ) { $chain = $$options{'chain'}; }
    }
    
    my $seq = '';
    my @seq = ();
    for my $residue (@$pdb)
    {
        if ( $chain && $chain ne $$residue{'chain'} ) { next }

        my $letter = $$residue{'one_letter'};
        if ( $upper_case ) { $letter = uc($letter); }
        
        if ( $output_array ) 
        {
            push @seq, $letter;
        }
        else
        {
            $seq .= $letter;
        }
    }

    if ( $output_array ) { return \@seq; }
    return $seq;
}


sub read_pdb_sequence
{
    my ($fname,$options) = @_;

    if ( $options && ref($options) ne 'HASH') { exit_nicely("FIXME: read_pdb_sequence iface changed.\n") }

    my $chain;
    my $output_array = 0;
    my $upper_case = 0;
    if ( ref($options) eq 'HASH' )
    {
        if ( exists($options->{array}) && $options->{array}==1 ) { $output_array = 1; }
        if ( exists($options->{uc}) && $options->{uc}==1 ) { $upper_case = 1; }
        if ( exists($options->{chain}) ) { $chain = $options->{chain}; }
    }

    my ($fh,$must_close);
    
    if ( ref($fname) eq 'GLOB' ) { $fh = $fname; }
    else { open($fh,"<$fname") || exit_nicely("$fname: $!"); $must_close=1; }

    my $prev_id = -1;
    my $seq = '';
    my @seq = ();
    while (my $line=<$fh>)
    {
        # ATOM      1  HN2 TYR     1      -0.338   0.478   0.828
        if ( !($line=~/^ATOM/) ) { next; }
        
        my $name       = substr($line,17,3);
        my $chain_id   = substr($line,21,1);
        my $seq_number = substr($line,22,5);

        if ( $chain && $chain_id ne $chain ) { next; }

        $name = lc($name);
        if ( ! exists($three_to_one->{$name}) ) { exit_nicely("FIXME: unknown 3-letter code: '$name'\n"); }
            
        $seq_number =~ s/\s+//g; 
        if ( $prev_id != $seq_number ) 
        {
            if ( $prev_id!=-1 && $seq_number!=$prev_id+1 )
            {
                print STDERR "Warning: Bad residues numbering. After $prev_id follows $seq_number.\n";
            }
            my $letter = $upper_case ? uc($three_to_one->{$name}) : $three_to_one->{$name};
            if ( $output_array ) 
            {
                push @seq, $letter;
            }
            else
            {
                $seq .= $letter;
            }
            $prev_id = $seq_number;
        }
    }
    
    if ( $must_close ) { close($fh); }

    if ( $output_array ) { return \@seq; }
    return $seq;
}


sub residue_distance
{
    my ($a,$b) = @_;

    my $na = scalar @$a;
    my $nb = scalar @$b;

    my $min_dist = -1;
    for (my $ia=0; $ia<$na; $ia++)
    {
        for (my $ib=0; $ib<$nb; $ib++)
        {
            my $dist = ($$a[$ia]->{x} - $$b[$ib]->{x})**2 + ($$a[$ia]->{y} - $$b[$ib]->{y})**2 + ($$a[$ia]->{z} - $$b[$ib]->{z})**2;
            if ( $min_dist==-1 || $dist<$min_dist ) { $min_dist = $dist; }
        }
    }
    return sqrt($min_dist);
}


sub get_pdb_part
{
    my ($pdb,$subseq,$options) = @_;

    my $sequence = get_sequence($pdb);
    my ($seq_start,$seq_end) = split(/\.\./,$subseq);

    my $from = index($sequence,$seq_start);
    if ( $from==-1 ) { exit_nicely("No such substring \"$seq_start\" in the sequence \"$sequence\".\n") }
    if ( index($sequence,$seq_start,$from+1) != -1 ) { exit_nicely("Multiple occurances of \"$seq_start\" in the sequence \"$sequence\".\n") }

    my $to = $from + length($seq_start);
    if ( $seq_end )
    {
        $to = index($sequence,$seq_end);
        if ( $to==-1 ) { exit_nicely("No such substring \"$seq_end\" in the sequence \"$sequence\".\n") }
        if ( index($sequence,$seq_start,$to+1) != -1 ) { exit_nicely("Multiple occurances of \"$seq_end\" in the sequence \"$sequence\".\n") }
        $to += length($seq_end);
    }

    my @ret = @$pdb;
    splice(@ret,$to);
    splice(@ret,0,$from);
    return \@ret;
}



sub strip_pdb
{
    my ($pdb,$options) = @_;

    if ( !$options || ref($options) ne 'HASH') { exit_nicely("FIXME: strip_pdb iface changed.\n") }

    my $chain;
    my $strip;
    if ( exists($$options{'chain'}) ) { $chain = $$options{'chain'}; }
    if ( exists($$options{'strip'}) ) { $strip = $$options{'strip'}; }

    my @pdb_chain;
    if ( $chain )
    {
        for my $residue (@$pdb)
        {
            if ( $$residue{'chain'} ne $chain ) { next; }
            push @pdb_chain, $residue;
        }
    }
    else
    {
        @pdb_chain = @$pdb;
    }

    my @result;
    if ( $strip )
    {
        my @strips = split(/,/,$strip);
        for my $strip (@strips)
        {
            my $pdb_part = PDB::get_pdb_part(\@pdb_chain,uc($strip),$options);
            push @result,@$pdb_part;
        }
    }
    else
    {
        @result = @$pdb;
    }

    return \@result;
}



sub fix_numbering
{
    my ($pdb) = @_;

    my $new_to_old = {};
    my $ires  = 1;
    my $iatom = 1;
    for my $res (@$pdb)
    {
        $res->{res_number} = $ires;
        my $atoms = $res->{atoms};
        for my $atom (@$atoms)
        {
            $atom->{res_number} =~ s/\s*//g;
            $atom->{atom_id}    =~ s/\s*//g;
            $new_to_old->{"$res->{res_id}:$atom->{atom_id}"} = "$ires:$iatom";
            $atom->{res_number} = $ires;
            $atom->{atom_id}    = $iatom;
            $iatom++;
        }
        $ires++;
    }
    return $new_to_old;
}



# MolMol does not recognise the chain IDs listed in PDB, it works with
#   sequential numbering instead. The subroutine returns a hash, the keys
#   are the chain IDs and the values are their numbers, indexed from 0.
#
sub get_chain_numbering
{
    my ($pdb) = @_;

    my $numbering = {};
    my $chain_idx = 0;
    my $prev_chain;
    for my $resi (@$pdb)
    {
        my $chain = $$resi{'chain'};
        if ( ! defined $chain ) { use Data::Dumper; print STDERR Dumper($resi); }
        if ( $prev_chain && $chain eq $prev_chain ) { next }
        $prev_chain = $chain;
        if ( exists($$numbering{$chain}) ) { next } # Ignore misbehaved chains
        $$numbering{$chain} = $chain_idx++;
    }
    return $numbering;
}


1;

