
package motifmaker; 
use Exporter;
@ISA=('Exporter');         #profile names
@EXPORT=qw (&makeprofile &fullprofile2motif &motiflist2profile
            &autoevalparam &adjuststraymotif &IterateR);


$significant_motifs = 1;    # Never use the values directly - may change.
$similar_motifs     = 2;


#************************************************************************************************************
#
#SUBROUTINE                  'makeprofile'    		
#FUNCTION                    To make a global profile of a multiple alignment
#INPUT				        Array reference to sequence name and sequence strings
#OUTPUT                      Reference to global profiles that has headers.
#ALGORITHMS                  PSEUDOCODE
#                            Read sequence string and name
#                            Iterate each position of the multiple alignment and obtain AA in a column
#                            Convert each of the AA code in to PCP descriptors or vectors 1-5
#                            Calculate average, standard deviation and relative entropy
#                            Calculate the number of gaps, position and residue index of the first residue
#                            Print profile [GLOBAL]
#
#
# NOTE: the 21 element is the total entropy calculated using either the scaled sum approach
#   or a maximum of relative entropies. In the first case, the relative entropies will be
#   scaled.
#
# Rewrote the code to be able to work with any number of property values and any number of bins.
#
sub makeprofile 
{
    my ($alignment,$least_entrps,$least_smls) = @_;

    my @profile = ();
    my $len = length($$alignment[0]->{seq});
    my $prf_idx = 0;
    my $ref_idx = 0;

    for ($i=0; $i<$len; $i++)        # Absolute position in the alignment
    { 
        # If $include_gaps is set to 1, also gaps will be included.
        #    The gap may be only in the first sequence of the multiple alignment, therefore
        #    it is meaningful to include all columns.
        #
        my $consensus_letter = substr($$alignment[0]->{seq},$i,1);
        my $histogram = {};
        if ( $include_gaps || $consensus_letter ne '-')         # If the first (reference) sequence is not a gap
        {
            for my $seq (@$alignment)
            {
                my $aa = substr($seq->{seq},$i,1);
                $histogram->{$aa}++;
            }

            my ($entrps,$avgs,$devs,$total_entropy,$similarity,$pcdist) = Alignments::calc_entropy($histogram);

            # See also Alignments::calc_entropy for this.
            my $is_conserved = 1;
            for my $iprop (0..$Vectorlib::nprops-1)
            {
                if ( $$devs[$iprop]>1e-10 ) { $is_conserved=0; last }
            }

            # The least entropies are rounded to six digits.
            if ( scalar @$least_entrps && $$least_entrps[$i]+1e-5 < $total_entropy )
            {
                $total_entropy = $$least_entrps[$i];
                $is_conserved  = 0;
            }

            # The least similarities are rounded to six digits.
            if ( scalar @$least_smls && $$least_smls[$i]+1e-5 < $similarity )
            {
                $similarity = $$least_smls[$i];
            }

            $ref_idx++ unless $consensus_letter eq '-'; # Position relative to the reference sequence. Used in slist only.
            $prf_idx++;   # Index relative to the profile. When gaps are allowed, should be equal to abs_column. 
                          #     When gaps are disallowed, should be equal to the reference sequence position.

            my $position = 
            { 
                aln_column  => $i+1, 
                nseqs       => scalar @$alignment, 
                prf_line    => $prf_idx, 
                ref_idx     => $ref_idx,    # Used in .slist only.
                reference   => $consensus_letter, 
                avgs        => $avgs,
                devs        => $devs,
                entrps      => $entrps,
                total_entrp => $total_entropy,
                conserved   => $is_conserved,
                similarity  => $similarity,
                PCdist      => $pcdist
            };
            push @profile, $position;
        }
    }
    return \@profile;                 #return profile as a reference
}



#***************************************************************************************************************
#
#SUBROUTINE                 'fullprofile2motif'
#FUNCTION                   Convert global profile into motif list using RGL
#INPUT				       Array reference to global profile, R cutoff, G cutoff, L cutoff
#OUTPUT                     Array reference to Motif list
#ALGORITHMS                 PSEUDOCODE
#                           Read input global profile, R, G and L.
#                           Mark all singificant datalines{profile} that have relative entropy greater than or equal to R
#                           Calculate residue index difference [4th column] for two sequential singificant profiles
#                           If the difference is less than G then concatenate to the motif.
#                           If the difference is greater than G then start a new motif.
#                           Calculate the size of each motif (end res. position - start res position). If it is lesser than L
#                           Then add to stray motif list
#
sub fullprofile2motif 
{
    my ($inputarray, $R,$G,$L) = @_;

    my (@signif,@motiflist,@elementcount);

    if ( !defined($motif_type) ) { Exit::exit_nicely("Error: The motif type is not defined. Did you forget to specify the -t option?\n") }

    for my $position (@$inputarray)
    {
        if ( $motif_type==$significant_motifs )
        {
            if ( $$position{total_entrp} >= $R ) { push @signif, $$position{prf_line}; }
        }
        elsif ( $motif_type==$similar_motifs )
        {
            if ( $$position{similarity} >= $R ) { push @signif, $$position{prf_line}; }
        }
    }

    my @stray  = ();
    my @motifs = ();
    my @motif  = ();
    for (my $i=0; $i<=$#signif; $i++)  #iterate over the array containing significant residues
    { 
        push @motif, $signif[$i];

        # Calculate the residue distance between two significant residues. If the distance
        #   is too large, start a new motif.
        if ( $i+1<scalar @signif && ($signif[$i+1]- $signif[$i]) > ($G+1))      
        {
            if ( scalar @motif >= $L ) 
            { 
                if ( ! scalar @motif ) { die "FIXME: 1\n"; }
                push @motifs, [@motif];
            }
            else
            {
                push @stray, @motif;
            }
            @motif = ();
        }
    }
    if ( scalar @motif ) 
    { 
        if ( scalar @motif >= $L ) 
        { 
            if ( ! scalar @motif ) { die "FIXME: 2\n"; }
            push @motifs, [@motif];
        }
        else
        {
            push @stray, @motif;
        }
    }
    #push @motifs, [@stray];

    my @out = ();
    for $motif (@motifs)
    {
        push @out, { 'motif'=>$motif, 'R'=>$R }
    }

    return ("#PARAM R_G_L:$R:$G:$L:", \@out);
}


#***************************************************************************************************************
#*
#SUBROUTINE                  'adjuststraymotif'
#FUNCTION                     Adjusts stray motifs that were not included as a motif in the first pass
#INPUT				         Reference to motif list, New G cut off that can be allowed
#OUTPUT                       Modified motif list
#ALGORITHMS                   PSEUDOCODE
#                             Read in the motif list and new G
#                             Calculate difference of residue between first and last residues of each motif and a member of stray motiflist
#                             If the difference is less than G then add to motif (either first or last) where the difference is less
#                             Return reference to new motiflist
#
#VARIABLES             NAME							DESCRIPTION
#                      $inputlist                 Array reference to input motif list
#                      $allowgap                  New G-cutoff value
#                      @strayelem                 Array containing stray elements
#                      $i                         Array index for stray elements
#                      $min                       Arbitray number set to 1000 to reduce and find min
#                      $j                         Array index for motif list
#                      @storeA                    Stores all residue positions for each motif
#                      $fit                       Absolute difference between end or first residue and stray element
#                      $point                     String containing a flag to indicate whether the stray element is close to the start (0) or end (1)
#                      @tempA                     Array containing element of flag and stray element and motif
#                      $stray                     New stray motif string after adjusting
#                      @newlist                   Array of new motif list
#
#VERSION HISTORY
#        Current         Version 1.0
#        Release date    01MAR03
#        Written by      Venkatarajan S. Mathura
#
#***************************************************************************************************************
#
#
#* CALL:
#* $output1 = &adjuststraymotif(\@input1,$input2);
#* @input1  : Array of motif N-list.        Type:String Array
#* $input2  : G-cutoff                      Type:Number   #number of gaps allowed
#* $output1 : Array reference to new N-list.
#*

sub adjuststraymotif{

     my ($inputlist,$allowgap,@newlist);
     $inputlist=$_[0];  #input reference list
     @newlist=@$inputlist; #array after de-reference
     $allowgap=$_[1]+1; #number of gaps allowed between head or tail residues of the motifs
     my(@strayelem,$stray,@tempA,$i,$j);
     @strayelem=split(/\*/,$$inputlist[$#$inputlist]); #stray elements


    for($i=0;$i<=$#strayelem;$i++){    #stray elements  counter
       	$min=1000;                     #initial distance set to large number
     	for($j=1;$j<$#$inputlist;$j++){    #iterator over input list
     		@storeA=split(/\*/,$$inputlist[$j]);  #split the list based on *
           $fit=abs($storeA[0]-$strayelem[$i]);   # calculate distance between head and tail
           if($fit < $min){                       # check whether the distance is minimum
               $min=$fit;                         # if so set new minimum
               $point=$i."*".$j."*0";             # 0 for attaching to head
             }
            $fit=abs($storeA[$#storeA]-$strayelem[$i]);
             if($fit < $min){
               $min=$fit;
               $point=$i."*".$j."*1";             # 1 for attaching to tail
             }
        }
       @tempA=split(/\*/,$point);                 #split all points and regroup to create a N-motif list
       if($min <=$allowgap){
          if(substr($point,length($point)-1,1) eq "0"){
          	$newlist[$tempA[1]]= $strayelem[$tempA[0]]."*". $newlist[$tempA[1]];
          }elsif(substr($point,length($point)-1,1) eq "1"){
          	$newlist[$tempA[1]]= $newlist[$tempA[1]]."*".$strayelem[$tempA[0]];
          }
       }else{
          $stray.=$strayelem[$tempA[0]]."*";   #rest push it into stary list
       }
    }
    $newlist[0]=$newlist[0]." #STRAY ADJUST:".$_[1].":"; #Mark what you adjusted in the new list
    $newlist[$#newlist]=substr($stray,0,length($stray)-1);
	return \@newlist;
}



sub IterateR 
{
    my ($inputprofile, $RFrom, $RTo, $RStep, $gvalue,$lvalue) = @_;
    my @Gvalues=($gvalue);
    my @Lvalues=($lvalue);
	    
    my ($tempG,$tempL,$tempres);
    my ($Motif);
    my (@temp,$from,$count,@scheme,@FirstSeq,$PositionFirstSeq,$MotifEntropyList);

    # Get the first squence in the globle profile 
    for my $pos (@$inputprofile) 
    { 
        push @FirstSeq, $$pos{'reference'};
    }
    my $len  = scalar @FirstSeq;
    push @scheme, '        ';
    push @scheme, '        ';
    push @scheme, '        ';
    for (my $i=0; $i<$len; $i++)
    {
        $scheme[0] .= int(($i+1) / 100) % 10;
        $scheme[1] .= int(($i+1) / 10) % 10;
        $scheme[2] .= ($i+1) % 10;
    }
    push @scheme, '        '.(join('',@FirstSeq));
	
    my @MotifCandidates = ();   # store the candidate of Motif


    # Sanity check, we do not want the PCPMer website being tricked into
    #   infinite loops.
    if ( $RFrom > $RTo ) { Exit::exit_nicely("Error: The variable range parameter must go in the order min,max,step.\n") }
    if ( $RStep <= 0 ) { Exit::exit_nicely("Error: The variable step must be bigger than 0.\n") }

    my $prev_tempr;
    for (my $tempR=$RFrom; $tempR<=$RTo; $tempR+=$RStep)
    {
        if ( defined($prev_tempr) && $tempR==$prev_tempr )
        {
            Exit::exit_nicely("Error: Infinite loop? Please check the -v arguments, it seems that the step was chosen too low? ($RFrom,$RTo,$RStep)\n");
        }
        $prev_tempr = $tempR;

        foreach my $tempG (@Gvalues)
        {
            foreach my $tempL (@Lvalues)
            {
                my ($info,$tempmotifs) = fullprofile2motif($inputprofile,$tempR,$tempG,$tempL);
                $from  = 1;
                $count = 0;
                $line  = sprintf "%4.2f    ",$tempR;
                $PositionFirstSeq = 0;
                for (my $i=0; $i<scalar @$tempmotifs; $i++)
                {
                    my @temp = @{$$tempmotifs[$i]->{'motif'}};
                    if ( ! scalar @temp ) { next }

                    #print STDERR "motif $i .. ",join(' ',@temp),"\n";
                    my $motif = 
                    {
                        R    => $tempR,
                        from => $temp[0],
                        to   => $temp[$#temp],
                        seq  => \@temp
                    };
                    push(@MotifCandidates,$motif);

                    for (my $j=$from; $j<$temp[0]; $j++)
                    {
                        $line .= '-';
                        $PositionFirstSeq++;
                    }
                    for (my $j=$temp[0]; $j<=$temp[$#temp]; $j++)
                    {
                        $line .= $FirstSeq[$PositionFirstSeq];
                        $PositionFirstSeq++;
                    }

                    $count++;
                    $from = $temp[$#temp]+1;
                }
                while ($PositionFirstSeq < scalar @FirstSeq)
                {
                    $line .= '-';
                    $PositionFirstSeq++;
                }
                push(@scheme,$line);
            }
        }
    }
    
    # Now add two lines indicating which positions were totally conserved. This we
    #   learn from the average deviation fields calculated in Alignments::calc_entropy.
    #
    my $line1 = '        ';
    my $line2 = '        ';
    for my $pos (@$inputprofile)
    {
        if ( $$pos{conserved} )
        {
            $line1 .= $$pos{reference};
            $line2 .= '*';
        }
        else
        {
            $line1 .= ' ';
            $line2 .= ' ';
        }
    }
    push @scheme,$line2,$line1;

    # Go through the motif candidates and see if some of the motifs
    #   cannot be shortened.
    #
    my @motifs = ();
    for (my $j=0; $j<scalar @MotifCandidates; $j++) 
    {
        my $candidate     = $MotifCandidates[$j];
        my $FromCandidate = $candidate->{from};
        my $ToCandidate   = $candidate->{to};
        my $shorter_variant_exists = 0;
        for (my $i=$j+1 ;$i<scalar @MotifCandidates; $i++)
        {
            my $From = $MotifCandidates[$i]->{from};
            my $To   = $MotifCandidates[$i]->{to}; 
            if (($FromCandidate <= $From) &&  ($ToCandidate >= $To))
            {
                $shorter_variant_exists = 1;	
                last;
            }
        }
        if (!$shorter_variant_exists)
        {
            push(@motifs, $candidate);
        }
    }
    @motifs = sort { return ${$a->{seq}}[0]<=>${$b->{seq}}[0]; } @motifs;

    my @motiflist;
    my $info = "#PARAM R_G_L: 0 ".scalar @motifs." ";
    for my $motif (@motifs)
    {
        #$info .= sprintf " %.2f", $motif->{R};
        push @motiflist, { 'motif'=>$$motif{'seq'}, 'R'=>$$motif{'R'} };
    }
    $info .= ' :x:x:';
    #push @motiflist, [1,2];     # Not very nice - for compatability print_slist and print_profile

    return ($info,\@motiflist, \@scheme);	
}


1;

