/*
 * The logic of this program follows the paper
 *        Venkatarajan, M.S., Schein, C. H., Braun W., 2003, 
 *        "Identifying physical chemical property based sequence motifs in protein
 *        families and superfamilies: Application to DNase I related endonuclease"
 *        Bioinformatics 19:1381-1390
 * 
 *  The Bayesian method (Eq.14) is implemented in the motifsearch script, from
 *  which is this program called. In this program, the Eqs. 2-6 are implemented.
 *
 *  The format of the produced output file is 
 *      0       0       105     6.215852e-01
 *      |       |        |      +--- Score
 *      |       |        +---------- Offset in the sequence (starting from 0)
 *      |       +------------------- Motif number (starting from 0)
 *      +--------------------------- Index of the sequence in the database or alignment (starting from 0)
 *
 *  By default, only the offset with maximum score is printed. With the -v option, all
 *  offsets will be reported. Use with caution, this produces huge output!
 *  
 *  No check about the sanity of the database format is performed. This should
 *  be done in motifsearch. Currently it is not checked at all, the fasta format 
 *  is assumed. (The gaps are not checked as well.)
 */
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <glib.h>
#include <ctype.h>
#include <errno.h>
#include "conf.h"

#define STATUS_OK  0
#define STATUS_ERR 1


#define AA(x) ((unsigned short int)(x))
typedef struct
{
    GHashTable *config;
    int nprops;                 // Number of descriptors (traditionally E1-E5)
    int naacids;                // The size of the translation table.
    int *aacids;                // Translation table: acids[AA('R')] -> index in vectors.
    float **vectors;            // Access the quantitative descr. q of the aminoacid i as vectors[q][i], q=0..nprops-1, i=0..19
    GList *motifs;              // Each node points to a motif_t structure.
    FILE *out_fd;               // File descriptor for the results.
    float weight_devs;          // The weight W for standard deviation in Eq.2
    float denom_shift;          // The small positive phi shift added to denominator in Eq.2
    float signif_threshold;     // If relative entropy of a property for some residue in a motif is lower, it will be skipped.
    int ignore_shorter_than;    // Sequences shorter than this threshold (usually 30) will be ignored.
    int verbose;                // If 0, only offset with the maximum score will be reported. When non-zero, all offsets will be printed.
}
data_t;

typedef struct
{
    int   id;                   // Motif ID is an index to the .PCPprf file, numbered from 0.
    int   nres;                 // The number of residues in the motif.
    int  *idxs;                 // The relative offsets. (E.g. 0,1,3 for the residues indexed as 5,6,8.)
    char *aacids;               // The motif's amino acid one-letter codes.
    float *mean;                // The iprop-th property of the idx-th residue is accessible as mean[iprop*nres+idx], where
    float *devs;                //      iprop=0..nprops-1 and idx=0..nres-1
    float *entrps;              
}
motif_t;


// Utility functions
#define debug(x...) fprintf(stderr,x)
void error(const char *format, ...);
void free_list_and_data(GList *list);
char *get_nth_element(char *str, int n);

// Reading data
void init_property_vectors(data_t *data);
void read_motifs(data_t *data);

// Calculations
void find_scores(data_t *data);
float calc_score(data_t *data, motif_t *motif, char *sequence, size_t seqlen, int sequence_id, int motif_id);


int main(int argc, char **argv)
{
    data_t data;
    int i;

    data.config  = NULL;
    data.out_fd  = NULL;
    data.verbose = 0;

    // Read command line arguments
    for (i=1; i<argc; i++)
    {
        if ( !strcmp(argv[i],"-v") || !strcmp(argv[i],"--verbose")  )
        {
            data.verbose = 1;
            continue;
        }
        if ( !strcmp(argv[i],"-c") || !strcmp(argv[i],"--config")  )
        {
            if ( ++i>=argc ) error("Expected a file name after the %s parameter.\n", argv[i-1]);
            data.config = conf_read(argv[i]);
            continue;
        }
        if ( !strcmp(argv[i],"-o") || !strcmp(argv[i],"--out")  )
        {
            if ( ++i>=argc ) error("Expected a file name after the %s parameter.\n", argv[i-1]);
            data.out_fd = fopen(argv[i],"w");
            if ( !data.out_fd )
                error("%s: %s\n", argv[i],strerror(errno));
            continue;
        }
        if ( !strcmp(argv[i],"-h") || !strcmp(argv[i],"--help") || !strcmp(argv[i],"-?")  )
            error(NULL);

        error("Unknown argument \"%s\". Run with -h for help.\n", argv[i]);
    }

    if ( !data.config ) error("Missing the -c option.\n");
    if ( !data.out_fd ) error("Missing the -o option.\n");

    data.weight_devs = conf_get_float(data.config,"weight_devs",1.5);
    data.denom_shift = conf_get_float(data.config,"denom_shift",0.001);
    data.signif_threshold    = conf_get_float_or_die(data.config,"signif_threshold");
    data.ignore_shorter_than = conf_get_int(data.config,"ignore_shorter_than", 30);

    init_property_vectors(&data);
    read_motifs(&data);
    find_scores(&data);

    fclose(data.out_fd);

    return STATUS_OK;
}


void error(const char *format, ...)
{
    va_list ap;

    if ( !format )
    {
        fprintf(stderr,
                "Usage: motifminer [OPTIONS]\n"
                "Options: \n"
                "   -c, --config <file>\n"
                "   -o, --out <file>\n"
                "   -v, --verbose           Output scores for all offsets, not just the maximum score.\n"
               );
        exit(STATUS_ERR);
    }

    if (format)
    {
        va_start(ap, format);
        vfprintf(stderr,format,ap);
        va_end(ap);
    }
    exit(STATUS_ERR);
}

void free_list_and_data(GList *list)
{
    GList *l = list;
    while (l)
    {
        free(l->data);
        l = l->next;
    }
    g_list_free(list);
}



void init_property_vectors(data_t *data)
{
    GList *values, *l;
    int i, iprop, nvals;
    char c, buf[10];

    data->nprops  = conf_get_int_or_die(data->config, "nprops");
    data->vectors = malloc(sizeof(float*)*data->nprops);
    data->naacids = 123;
    data->aacids  = malloc(sizeof(int)*data->naacids);

    for (i=0; i<data->naacids; i++)
        data->aacids[i] = -1;

    l = values = conf_get_list_or_die(data->config, "order",',');
    nvals = 0;
    while (l)
    {
        sscanf(l->data,"%c",&c);
        if ( AA(c)>=data->naacids ) 
            error("FIXME: Some of the amino acids is out of range: \"%s\".\n", conf_get_string_or_die(data->config,"order"));

        data->aacids[AA(toupper(c))] = nvals;
        data->aacids[AA(tolower(c))] = nvals;

        l = l->next;
        nvals++;
    }

    free_list_and_data(values);
    
    for (iprop=0; iprop<data->nprops; iprop++)
    {
        snprintf(buf,10,"%d",iprop+1);
        values = conf_get_list_or_die(data->config,buf,',');
        data->vectors[iprop] = malloc(sizeof(float)*data->naacids);

        i = 0;
        l = values;
        while (l)
        {
            if ( i>nvals )
                error("Expected same number of values for \"%s\" as there were for \"order\".\n", buf);

            if ( sscanf(l->data,"%f",&(data->vectors[iprop][i]))!=1 )
                error("Could not parse the value \"%s\", expected a float.\n", l->data);

            l = l->next;
            i++;
        }
        free_list_and_data(values);
    }
}


void read_motifs(data_t *data)
{
    int iprop, slen, avgcol_from=-1,avgcol_to=-1;
    int motiflen, idx, done,imotif;
    char *lineptr = NULL, *tmp;
    size_t linelen = 0;
    FILE *fd;
    motif_t *motif = NULL;
    char *prf_file = conf_get_string_or_die(data->config,"motifs");
    
    fd = fopen(prf_file,"r");
    if ( !fd )
        error("%s: %s\n", prf_file,strerror(errno));

    data->motifs = NULL;

    imotif = 0;
    done = 0;
    slen = strlen("#COLUMNLABELS");
    do
    {
        if ( getline(&lineptr, &linelen, fd) <=0 )
            done = 1;

        // #MOTIF NUM_SIGCOUNT:2:8:
        if ( done || !strncmp("#MOTIF NUM_SIGCOUNT", lineptr, slen) )
        {
            if ( motif )
            {
                if ( motif->nres != motiflen )
                    error("The motiflen given in the header (%d) differs from the real motif length (%d).\n",motiflen,motif->nres);

                // For each residue of a motif create relative numbering counted from 0. That is, if a motif
                //  constists of residues 6,7,8 -> 0,1,2.
                for (idx=motiflen-1; idx>=0; idx--)
                    motif->idxs[idx] -= motif->idxs[0];  
                
                if ( !conf_get_string(data->config,"motifs_list",NULL)  || conf_is_int_in_list(data->config,"motifs_list", imotif) )
                {
                    data->motifs = g_list_append(data->motifs,motif);
                    motif = NULL;
                }

                imotif++;
            }

            if ( done )
                break;

            if ( 1!=sscanf(lineptr, "#MOTIF NUM_SIGCOUNT:%*d:%d", &motiflen) )
                error("Could determine the motif length: %s\n", lineptr);

            if ( !motif )
                motif = malloc(sizeof(motif_t));

            motif->id     = imotif;
            motif->nres   = 0;
            motif->idxs   = malloc(sizeof(int)*motiflen);
            motif->aacids = malloc(sizeof(int)*motiflen);
            motif->mean   = malloc(sizeof(float)*data->nprops*motiflen);
            motif->devs   = malloc(sizeof(float)*data->nprops*motiflen);
            motif->entrps = malloc(sizeof(float)*data->nprops*motiflen);

            continue;
        }

        // #COLUMNLABELS POS[1] TOTSEQ[2] SEQNOGAP[3] RESINSEQ1[4] RESCODE[5] AVG[6-10] STD[11-15] R[16-20]
        if ( !strncmp("#COLUMNLABELS", lineptr, slen) )
        {
            if ( 2!=sscanf(lineptr,"#COLUMNLABELS POS[1] TOTSEQ[2] SEQNOGAP[3] RESINSEQ1[4] RESCODE[5] AVG[%d-%d]", &avgcol_from,&avgcol_to) )
                error("Could not parse the line: %s\n", lineptr);

            if ( avgcol_to-avgcol_from+1 != data->nprops )
                error("The number of columns (%d) in %s does not correspond to the number of descriptors given in the config file (%d).\n", 
                    avgcol_to-avgcol_from+1, prf_file,data->nprops);

            slen = strlen("#MOTIF NUM_SIGCOUNT");
            continue;
        }

        if ( !motif )
            continue;

        idx = motif->nres++;
        if ( idx>=motiflen )
            error("The motiflen given in the header (%d) is smaller than the real motif length.\n", motiflen);

        //  21   7   7  21 Q   8.8769  -7.0910  -5.4263   2.0313   2.8814   1.4406   1.2175   2.9911   0.9166   2.9292   0.8595   1.5524   0.9254   2.6430   1.9394
        // Position tmp to the 4th column, that is, skip 3 elements.
        tmp = get_nth_element(lineptr,4);
        if ( 1!=sscanf(tmp,"%d",&(motif->idxs[idx])) )
            error("Could not parse the position index in %s\n", lineptr);

        // Move tmp to the 5th column, that is, skip 1 element.
        tmp = get_nth_element(tmp,2);
        if ( 1!=sscanf(tmp,"%c",&(motif->aacids[idx])) )
            error("Could not parse the amino acid letter in %s\n", lineptr);

        // Average properties
        // Move tmp to the 6th column, skip 1 element.
        tmp = get_nth_element(tmp,2);
        for (iprop=0; iprop<data->nprops; iprop++)
        {
            // Read the element, do not move tmp, only to eat spaces.
            tmp = get_nth_element(tmp,1);
            if ( !tmp ) 
                error("Could not determine the %d mean in %s\n", iprop+1,lineptr);
            if ( 1!=sscanf(tmp,"%f",&(motif->mean[iprop*motiflen+idx])) )
                error("Could not parse the %d mean in %s\n, iprop+1,lineptr");

            // Skip the element we've just read.
            tmp = get_nth_element(tmp,2);
        }

        // Standard deviations
        for (iprop=0; iprop<data->nprops; iprop++)
        {
            tmp = get_nth_element(tmp,1);
            if ( !tmp ) 
                error("Could not determine the %d dev in %s\n", iprop+1,lineptr);
            if ( 1!=sscanf(tmp,"%f",&(motif->devs[iprop*motiflen+idx])) )
                error("Could not parse the %d dev in %s\n, iprop+1,lineptr");
            tmp = get_nth_element(tmp,2);
        }

        // Relative entropies
        for (iprop=0; iprop<data->nprops; iprop++)
        {
            tmp = get_nth_element(tmp,1);
            if ( !tmp ) 
                error("Could not determine the %d entrp in %s\n", iprop+1,lineptr);
            if ( 1!=sscanf(tmp,"%f",&(motif->entrps[iprop*motiflen+idx])) )
                error("Could not parse the %d entrp in %s\n, iprop+1,lineptr");
            tmp = get_nth_element(tmp,2);
        }

        // Eat also the rest of fields and check that we get null at the right moment.
        // Total entropy, similarity, physicochemical distance
        tmp = get_nth_element(tmp,3);
        if ( !tmp )
            error("The profile contains less columns than expected: %s\n", prf_file);
        tmp = get_nth_element(tmp,2);
        if ( tmp )
            error("The profile contains more columns than expected: %s\n", prf_file);
    }
    while ( !done );
    fclose(fd);

    if ( !data->motifs )
        error("No motifs read from %s.\n", prf_file);
}


// The index n is counted from 1. If the string starts with spaces, the first non-space element
//  is still the 1st one, unlike the "cut" command line utility.
//  If str is positioned at non-whitespace character and n is 1, str will be returned unchanged.
//
char *get_nth_element(char *str, int n)
{
    char *tmp = NULL;
    int ielement=0;

    if ( !str ) return NULL;

    do
    {
        if ( tmp ) str = tmp;

        while (isspace(*str))
            str++;

        if ( !*str ) return NULL;   // This should not happen

        tmp = str;                  // Find the next space
        while (*tmp && !isspace(*tmp))
            tmp++;

        ielement++;
    }
    while (ielement<n);

    return str;
}


// Read the database line by line and calculate scores. Multilines are
//  stored in buf.
//
void find_scores(data_t *data)
{
    FILE *db_fd=NULL;
    char *lineptr=NULL, *buf=NULL;
    int done, iseq;
    size_t linelen=0, bufused=0, bufsize=0, len;
    GList *l;
    motif_t *motif;

    // THis is added by negi for allergen project
    // PCPmer database files are random here.. I guess..
    // sequence no from database are different from as observed in output score.
    // this is added so that we can compate ZSOCRE correctly
    FILE *db_junk;
    db_junk = fopen("JUNK.dat","a");

    db_fd = fopen(conf_get_string_or_die(data->config,"database"),"r");
    if ( !db_fd )
        error("%s: %s\n", conf_get_string_or_die(data->config,"database"), strerror(errno));

    iseq = 0;
    done = 0;
    do
    {
        if ( getline(&lineptr, &linelen, db_fd) <=0 )
            done = 1;

        if ( done || *lineptr == '>' )
        {
	    // Print what file header is comming first..
	    // Added by negi
	    // this will write a JUNK.dat file in input dir..
    	    fprintf(db_junk,"%s", lineptr);

            if ( !buf ) continue;

            l = data->motifs;
            while (l)
            {
                if ( data->verbose || bufused >= data->ignore_shorter_than )
                {
                    motif = (motif_t*) l->data;
                    calc_score(data, l->data, buf, bufused, iseq,motif->id);
                }
                l = l->next;
            }

            buf[0]  = 0;
            bufused = 0;
            iseq++;

            continue;
        }

        len = strlen(lineptr);
        while ( len>0 && isspace(lineptr[len-1]) )  // Leave out ending spaces and newline characters
            len--;
        if ( !len )
            error("FIXME: this should not happen, the DB contains an empty line!\n");

        if ( len+bufused >= bufsize )   // The terminating null byte is not counted by bufused.
        {
            bufsize += len*2;           // Allocate larger chunk than actually needed. (Fewer malloc calls.)
            buf = realloc(buf,bufsize);
        }
        strncpy(buf+bufused, lineptr, len);
        bufused += len;
        buf[bufused] = 0;
    }
    while (!done);

    fclose(db_fd);

    if ( buf )
        free(buf);
    if ( lineptr )
        free(lineptr);
}


float calc_score(data_t *data, motif_t *motif, char *sequence, size_t seqlen, int sequence_id, int motif_id)
{
    int win_pos, end_pos, iseq, iaa, max_score_win_pos=0;
    int ires, count;
    int iprop;
    float score, max_score=0, zvalue;

    end_pos = seqlen - motif->idxs[motif->nres-1];

    for (win_pos=0; win_pos<end_pos; win_pos++)
    {
        score = 0;
        count = 0;
        for (ires=0; ires<motif->nres; ires++)
        {
            iseq = win_pos + motif->idxs[ires];
            iaa  = data->aacids[AA(sequence[iseq])];

            for (iprop=0; iprop<data->nprops; iprop++)
            {
                if ( motif->entrps[iprop*motif->nres+ires] < data->signif_threshold ) continue;
                    
                zvalue  = data->vectors[iprop][iaa] - motif->mean[iprop*motif->nres+ires];
                zvalue /= data->denom_shift + motif->devs[iprop*motif->nres+ires] * data->weight_devs;
                score  += 1.0/(1+zvalue*zvalue);
                count++;
            }
        }
        score /= count;
	// negi.. this file is dumped in to and then parse by motifsearh
        if ( data->verbose )
            fprintf(data->out_fd,"%d\t%d\t%d\t%e\t%c\n", sequence_id,motif_id,win_pos,score, sequence[win_pos]);

        if ( max_score < score )
        {
            max_score = score;
            max_score_win_pos = win_pos;
        }
    }

    if ( ! data->verbose )
        fprintf(data->out_fd,"%d\t%d\t%d\t%e\n", sequence_id,motif_id,max_score_win_pos,max_score);

    return max_score;
}


