package Utils;


# Given full path, creates the output directory and returns a prefix
#   valid for all output files. Use as 
#
#   my $prefix = init_result_dir("/path/to/some/dir");
#   my $fname  = "$prefix.my_suffix";
#
# Tested with: 
#   -p results/dv/      ..  results/dv/dv.*
#   -p results/dv       ..  results/dv.*
#   -p dv               ..  dv.*
#
sub init_result_dir
{
    my ($prefix) = @_;

    if ( !$prefix ) { $prefix='out/'; }
    if ( $prefix =~ m{/?([^/]*)$} )
    {
        my $dir = $`;
        my $file = $1;
        if ( !$file ) 
        { 
            $file = $dir;
            if ( $file=~ m{([^/]+)/?$} )
            {
                $file = $1;
            }
            else { Exit::exit_nicely("FIXME: Could not parse dir=$dir, file=$file.\n");}
        }
        if ( !$dir )
        {
            $dir = '.';
        }
        $prefix = "$dir/$file";
        if ( $dir && ! -e $dir ) { `mkdir -p $dir`; }
    }

    return $prefix;
}




# If you make any changes here, make sure that also Utils::read_prf_file is changed
#   accordingly. The residues in motifs are indexed from 1.
#
sub print_profile
{
    my ($fname,$info,$motifs,$global_prf) = @_;

    open(OUT,">$fname") or usage("$fname: $!\n");
    print OUT "$info\n";
    printf OUT "#COLUMNLABELS POS[1] TOTSEQ[2] SEQNOGAP[3] RESINSEQ1[4] RESCODE[5] AVG[6-%d] STD[%d-%d] Entrp[%d-%d] TotEntrp[%d] Similarity[%d] PCdist[%d]\n",
        5+$Vectorlib::nprops, 
        6+$Vectorlib::nprops, 5+2*$Vectorlib::nprops, 
        6+2*$Vectorlib::nprops, 5+3*$Vectorlib::nprops,
        6+3*$Vectorlib::nprops,
        7+3*$Vectorlib::nprops,
        8+3*$Vectorlib::nprops;


#
#	



    for (my $i=0; $i<scalar @$motifs; $i++)
    {
        my $motif = $$motifs[$i]; # $$motifs[$i-1];     # Ommit the last record - it is the stray motif.
        printf OUT "#MOTIF NUM_SIGCOUNT:%d:%d:%.2f\n",$i,scalar @{$$motif{'motif'}}, $$motif{'R'};
        for my $pos (@{$$motif{'motif'}})
        {
            if ( !exists($$global_prf[$pos-1]) ) { Exit::exit_nicely("FIXME: index out of bounds: $pos\n") }
            Utils::print_profile_line(\*OUT, $$global_prf[$pos-1]);
        }
    }
    close(OUT);
}



# Returned is a hash 
#   $motifs->{$motif_num}->{seq} = sequence, where $motif_num=0..nmotifs-1
#   @{$motifs->{$motif_num}->{residues}}
# 
# Note: the position indexes are left untouched, therefore they are indexed from 1
#
sub read_prf_file_motifs
{
    my ($fname) = @_;

    if ( !$Vectorlib::nprops ) { Exit::exit_nicely("Error: The Vectorlib must be included before the Utils::read_prf_file_motifs() is called.\n"); }

    open(FILE,"<$fname") || Exit::exit_nicely("$fname: $!\n");

    my $motifs = {};

    #   #PARAM R_G_L:2.1:2:4:
    #   #COLUMNLABELS POS[1] TOTSEQ[2] SEQNOGAP[3] RESINSEQ1[4] RESCODE[5] AVG[6-10] STD[11-15] R[16-20]
    #   #MOTIF NUM_SIGCOUNT:1:4:0.35
    #    21   7   7  21 Q   8.8769  -7.0910  -5.4263   2.0313   2.8814   1.4406   1.2175   2.9911   0.9166   2.9292   0.8595   1.5524   0.9254   2.6430   1.9394
    while (my $line=<FILE>)
    {
        if ( $line=~/^\#MOTIF NUM_SIGCOUNT:(\d+):(\d+):(\S+)?/ )
        {
            my $motifnum  = $1;
            my $nresidues = $2;
            my $R         = $3 ? $3 : 0;    # Backward compatibility with old profiles: remove at some point along with the ? mark in the regex above

            my $offset    = 0;
            my @residues  = ();
            for (my $i=0; $i<$nresidues; $i++)
            {
                $line = <FILE>;
                if ( !$line  ) 
                { 
                    Exit::exit_nicely("The file $fname ended prematurely, $nresidues lines expected, got $i for motif #$motifnum.\n"); 
                }
                my $prf_item = parse_profile_line($line);
                push @{$motifs->{$motifnum}->{'residues'}}, $prf_item;
                
                if ( $i==0 ) { $offset = $$prf_item{prf_line}; }
                $residues[$$prf_item{prf_line} - $offset] = $$prf_item{'reference'};
            }
            my $motif = '';
            for (my $i=0; $i<scalar @residues; $i++)
            {
                $motif .= exists($residues[$i]) ? $residues[$i] : '.';
            }
            $motifs->{$motifnum}->{'seq'} = $motif;
            $motifs->{$motifnum}->{'R'}   = $R;
        }
    }
    close(FILE);
    return $motifs;
}


# Returns a list of profile lines. See parse_profile_line().
#   Reconstructs the indexes relative to the reference sequence.
sub read_prf_file
{
    my ($fname) = @_;

    if ( !$Vectorlib::nprops ) { Exit::exit_nicely("Error: The Vectorlib must be included before the Utils::read_prf_file() is called.\n"); }

    open(FILE,"<$fname") || Exit::exit_nicely("$fname: $!\n");

    my $ref_idx = 0;
    my @profile = ();
    while (my $line=<FILE>)
    {
        my $items = parse_profile_line($line);
        if ( scalar keys %$items )
        {
            if ( $$items{'reference'} ne '-' ) 
            { 
                $$items{'ref_idx'} = ++$ref_idx;
            }
            else
            {
                $$items{'ref_idx'} = 0;
            }
            push @profile, $items;
        }
    }
    close(FILE);
    return \@profile;
}



# If changed, also parse_profile_line must be changed. Also make sure
#   that headers
#       #COLUMNLABELS POS[1] TOTSEQ[2] SEQNOGAP[3] RESINSEQ1[4] RESCODE[5] AVG[6-10] STD[11-15] R[16-20]
#   are changed accordingly: motifmaker::print_profile(), Utils::read_prf_file(). 
#
sub print_profile_line
{
    my ($fh,$item) = @_;

    if ( !$item ) { Exit::exit_nicely("FIXME: An attempt to write an empty profile line detected!\n") }

    #   21   7   7  21 Q   8.8769  -7.0910  -5.4263   2.0313   2.8814   1.4406   1.2175   2.9911   0.9166   2.9292   0.8595   1.5524   0.9254   2.6430   1.9394
    #   0    1   2  3  4   5
    #   Check the code for field names.
    #
    if ( !$Vectorlib::nprops ) { Exit::exit_nicely("Error: The Vectorlib must be included before the Utils::print_profile_line() is called.\n"); }

    my $fmt = "%3d %3d %3d %3d %s";
    for (1..$Vectorlib::nprops) { $fmt .= ' %8.4f' }    # avgs
    for (1..$Vectorlib::nprops) { $fmt .= ' %8.4f' }    # devs
    for (1..$Vectorlib::nprops) { $fmt .= ' %8.4f' }    # entrps
    $fmt .= " %8.4f %8.4f %8.4f\n";                     # total entropy, similarity, and avg physico-chemical distance

    printf $fh $fmt, 
        $$item{'aln_column'}, 
        $$item{'nseqs'}, 
        $$item{'nseqs'},
        $$item{'prf_line'},
        $$item{'reference'},
        @{$$item{'avgs'}},
        @{$$item{'devs'}},
        @{$$item{'entrps'}},
        $$item{'total_entrp'},
        $$item{'similarity'},
        $$item{'PCdist'};
}


sub parse_profile_line
{
    my ($line) = @_;

    if ( !$Vectorlib::nprops ) { Exit::exit_nicely("Error: The Vectorlib must be included before the Utils::parse_profile_line() is called.\n"); }

    my $info = {};

    if ( $line=~/^#/ ) { return $info }
    
    $line =~ s/^\s*//;
    my @items = split(/\s+/,$line);

    if ( scalar @items != $Vectorlib::nprops*3 + 8 )
    {
        Exit::exit_nicely(
                "Sanity check failed: Different number of columns in the profile line than expected\n",
                sprintf("(expected %d, found %d). Perhaps different vectorlib is used?\n",$Vectorlib::nprops*3 + 8, scalar @items),
                "The offending line was:\n\t",$line,"\n"
             );
    }

    $$info{'aln_column'} = shift(@items);
    $$info{'nseqs'}      = shift(@items); shift(@items);
    $$info{'prf_line'}   = shift(@items);
    $$info{'reference'}  = shift(@items);

    my @avgs = ();
    for my $iprop (1..$Vectorlib::nprops)
    {
        push @avgs, shift(@items);
    }
    $$info{'avgs'} = \@avgs;

    my @devs = ();
    for my $iprop (1..$Vectorlib::nprops)
    {
        push @devs, shift(@items);
    }
    $$info{'devs'} = \@devs;

    my @entrps = ();
    for my $iprop (1..$Vectorlib::nprops)
    {
        push @entrps, shift(@items);
    }
    $$info{'entrps'} = \@entrps;

    $$info{'total_entrp'} = shift(@items);
    $$info{'similarity'} = shift(@items);
    $$info{'PCdist'} = shift(@items);

    return $info;
}
       

sub format_html_motif_tables
{
    my ($motifs,$options) = @_;
    
    my $checkboxes = ( $options && exists($$options{'checkboxes'}) && $$options{'checkboxes'} ) ? 1 : 0;
    my $signif_th  = ( $options && exists($$options{'signif_threshold'}) ) ? $$options{'signif_threshold'} : 0;
    my $out = '';
    
    for my $imotif (sort {$a<=>$b} keys %$motifs)
    {
        my $prev_pos;
        my $residues = $$motifs{$imotif}->{'residues'};

        # The motifs contain gaps. First fill the gaps with empty
        #   elements, the gap-filling algorithm will be run only once
        #   this way.
        #
        my @motif = ();
        for (my $j=0; $j<scalar @$residues; $j++)
        {
            my $pos = $$residues[$j]->{'prf_line'};

            if ( $j==0 || $prev_pos+1==$pos )
            {
                push @motif, $$residues[$j];
                $prev_pos = $pos;
                next;
            }
            for (my $k=$prev_pos+1; $k<$pos; $k++)
            {
                push @motif, 0;
            }
            push @motif, $$residues[$j];
            $prev_pos = $pos;
        }
        
        $out .= "\n<table class='motif_table'><tr><td>";

        if ( $checkboxes )
        {
            $out .= "<input type='checkbox' name='motif_$imotif' value='$imotif'>";
        }
        for (my $j=0; $j<scalar @motif; $j++)
        {
            if ( $motif[$j] )
            {
                $out .= "<td title='$motif[$j]->{'aln_column'}'>$motif[$j]->{'reference'}";
            }
            else
            {
                $out .= "<td>.";
            }
        }
        $out .= "\n<tr><td>Simil";
        for (my $j=0; $j<scalar @motif; $j++)
        {
            if ( $motif[$j] )
            {
                my $simil = $motif[$j]->{'similarity'};
                my $color = Color::get_color_string("ff4444","ffffff","669966", $simil);
                $out .= sprintf "<td style='background-color:$color' title='%.3f'>", $simil;
            }
            else
            {
                $out .= "<td style='background-color:#ffffff'>";
            }
        }
        for my $iprop (0..$Vectorlib::nprops-1)
        {
            $out .= sprintf "\n<tr><td>D%d", $iprop+1;
            for (my $j=0; $j<scalar @motif; $j++)
            {
                if ( $motif[$j] )
                {
                    # The parameter 0.2 below is purely empirical to give visually pleasing results.
                    #   May be changed in the future. Tested for Flavivirus NS3 protein.
                    my $devs  = $motif[$j]->{'devs'};
                    my $dev   = $$devs[$iprop-1]==-1 ? 0 : exp(-$$devs[$iprop-1]*0.2);
                    my $color = Color::get_color_string("ffffff","ffffff","888888", $dev);
                    $out .= sprintf "<td style='background-color:$color' title='%.3f'>", $$devs[$iprop];
                }
                else
                {
                    $out .= "<td style='background-color:#ffffff'>";
                }
            }
        }
        $out .= "\n<tr><td>Tot E";
        for (my $j=0; $j<scalar @motif; $j++)
        {
            if ( $motif[$j] )
            {
                my $entrp = $motif[$j]->{'total_entrp'};
                my $color = $entrp < 0 ? "#ff4444" : Color::get_color_string("ff4444","ffffff","5555ff", $entrp);
                $out .= sprintf "<td style='background-color:$color' title='%.3f'>", $entrp;
            }
            else
            {
                $out .= "<td style='background-color:#ffffff'>";
            }
        }
        for my $iprop (0..$Vectorlib::nprops-1)
        {
            $out .= sprintf "\n<tr><td>E%d", $iprop+1;
            for (my $j=0; $j<scalar @motif; $j++)
            {
                if ( $motif[$j] )
                {
                    my $entrps = $motif[$j]->{'entrps'};
                    my $color  = $$entrps[$iprop] < 0 ? "#ffffff" : Color::get_color_string("ffffff","888888","000000", $$entrps[$iprop]);
                    $out .= sprintf "<td style='background-color:$color' title='%.3f'>", $$entrps[$iprop];
                    if ( $signif_th && $signif_th <= $$entrps[$iprop] )
                    {
                        $out .= 'x';
                    }
                }
                else
                {
                    $out .= "<td style='background-color:#ffffff'>";
                }
            }
        }
        $out .= "</table>\n";
    }
    return $out;
}

sub wrap_line
{
    my ($line,$options) = @_;

    my $length = 60;
    my $separator = "\n";
    if ( $options )
    {
        if ( exists($$options{'length'}) ) { $length=$$options{'length'}; }
        if ( exists($$options{'separator'}) ) { $separator=$$options{'separator'}; }
    }

    my $buffer;
    do
    {
        $buffer .= substr($line,0,$length,'');
        $buffer .= $separator;
    }
    while ($line);

    return $buffer;
}


1;
